/*
 * (UNISTIM*)
 *
 * UNISTIM channel driver for asterisk
 *
 * 2005 - cedric.hans@mlkj.net
 *
 * Version 0.9.4
 * 
 * This program is free software and may be modified and 
 * distributed under the terms of the GNU Public License.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/signal.h>
#include <asterisk/lock.h>
#include <asterisk/frame.h>
#include <asterisk/channel.h>
#include <asterisk/logger.h>
#include <asterisk/module.h>
#include <asterisk/pbx.h>
#include <asterisk/config.h>
#include <asterisk/options.h>
#include <asterisk/callerid.h>
#include <asterisk/features.h>
#include <asterisk/musiconhold.h>
#include <asterisk/say.h>
#include <asterisk/astdb.h>
#include <asterisk/utils.h>
#include <asterisk/cli.h>
#include <asterisk/app.h>
#include <asterisk/rtp.h>
#include <asterisk/acl.h>
#include <asterisk/dsp.h>

static const char desc[] = "UNISTIM for Asterisk";
static const char tdesc[] = "UNISTIM Channel Driver";
static const char type[] = "USTM";

static int usecnt = 0;

/* Beware, G729 and G723 are not supported by asterisk, except with the proper licence */
#define CAPABILITY AST_FORMAT_ALAW	/* | AST_FORMAT_ULAW  | AST_FORMAT_G729A | AST_FORMAT_G723_1 */
AST_MUTEX_DEFINE_STATIC (usecnt_lock);
/* This is the thread for the monitor which checks for input on the channels
 *    which are not currently in use.  */
static pthread_t monitor_thread = AST_PTHREADT_NULL;

/* Protect the monitoring thread, so only one process can kill or start it, and not
 *    when it's doing something critical. */
AST_MUTEX_DEFINE_STATIC (monlock);
/* Protect the network socket */
AST_MUTEX_DEFINE_STATIC (netlock);
/* Protect the session list */
AST_MUTEX_DEFINE_STATIC (sessionlock);
/* Protect the device list */
AST_MUTEX_DEFINE_STATIC (devicelock);

/************************************************************************************/
/*                            Asterisk specific globals                             */
/************************************************************************************/

static char context[AST_MAX_EXTENSION] = "default";
static char language[MAX_LANGUAGE] = "";
static char musicclass[MAX_LANGUAGE] = "";
static char callerid[AST_MAX_EXTENSION] = "";
static char default_callerid[AST_MAX_EXTENSION] = "asterisk";
static char default_callername[AST_MAX_EXTENSION] = "Unknown";
static char linelabel[AST_MAX_EXTENSION] = "";
static int nat = 0;
static unsigned int cur_callergroup = 0;
static unsigned int cur_pickupgroup = 0;
static int immediate = 0;
static int callwaiting = 0;
static int callreturn = 0;
static int threewaycalling = 0;
/* This is for flashhook transfers */
static int transfer = 0;
static int cancallforward = 0;
/* static int busycount = 3;*/
static char accountcode[20] = "";
static char mailbox[AST_MAX_EXTENSION];
static int amaflags = 0;
static int callnums = 1;


#define SUB_REAL 0
#define SUB_ALT  1
#define MAX_SUBS 2

//#define DUMP_PACKET 1
static void
dummy (char *dummy, ...)
{
  return;
}

//#define DEBUG_TIMER ast_verbose
#define DEBUG_TIMER dummy
int unistimdebug = 0;		// Enable verbose output. can also be set with the CLI 
int unistim_port;
int unistim_keepalive;
int taskid = -1;
unsigned int dwTimeOut = 0;
static struct sockaddr_in public_ip = { 0, };
static int unistimsock = -1;
static struct io_context *io;
AST_MUTEX_DEFINE_STATIC (unistim_reload_lock);
static struct sched_context *sched;
static int unistim_reloading = 0;
#define MAXINT		(unsigned int)(-1)
#define MAX_BUF_SIZE	64	// Size of the transmit buffer
#define MAX_BUF_NUMBER	30	// Number of slots for the transmit queue
#define NB_MAX_RETRANSMIT 8	// Try x times before removing the phone
#define IDLE_WAIT	1000	// Nb of milliseconds waited when no events are scheduled
#define RETRANSMIT_TIMER 2000	// Wait x milliseconds before resending a packet
#define DEFAULT_CODEC	0x00	// G711 Law

enum PhoneState
{
  STATE_INIT,
  STATE_AUTHDENY,
  STATE_MAINPAGE,
  STATE_DIALPAGE,
  STATE_DIALING,
  STATE_RINGING,
  STATE_CALL,
  STATE_SELECTCODEC,
  STATE_CLEANING
};

enum HandsetState
{
  STATE_ONHOOK,
  STATE_OFFHOOK,
};

enum PhoneKey
{
  KEY_0 = 0x40,
  KEY_1 = 0x41,
  KEY_2 = 0x42,
  KEY_3 = 0x43,
  KEY_4 = 0x44,
  KEY_5 = 0x45,
  KEY_6 = 0x46,
  KEY_7 = 0x47,
  KEY_8 = 0x48,
  KEY_9 = 0x49,
  KEY_STAR = 0x4a,
  KEY_SHARP = 0x4b,
  KEY_UP = 0x4c,
  KEY_DOWN = 0x4d,
  KEY_RIGHT = 0x4e,
  KEY_LEFT = 0x4f,
  KEY_QUIT = 0x50,
  KEY_COPY = 0x51,
  KEY_FUNC1 = 0x54,
  KEY_FUNC2 = 0x55,
  KEY_FUNC3 = 0x56,
  KEY_FUNC4 = 0x57,
  KEY_ONHOLD = 0x5b,
  KEY_HANGUP = 0x5c,
  KEY_MUTE = 0x5d,
  KEY_HEADPHN = 0x5e,
  KEY_LOUDSPK = 0x5f,
  KEY_FAV0 = 0x60,
  KEY_FAV1 = 0x61,
  KEY_FAV2 = 0x62,
  KEY_FAV3 = 0x63,
  KEY_FAV4 = 0x64,
  KEY_FAV5 = 0x65,
  KEY_COMPUTR = 0x7b,
  KEY_CONF = 0x7c,
  KEY_SNDHIST = 0x7d,
  KEY_RCVHIST = 0x7e,
  KEY_INDEX = 0x7f
};

typedef struct __WSABUF
{
  u_long len;
  char *buf;
}
WSABUF, *LPWSABUF;

typedef struct _SYSTEMTIME
{
  unsigned short wYear;
  unsigned short wMonth;
  unsigned short wDayOfWeek;
  unsigned short wDay;
  unsigned short wHour;
  unsigned short wMinute;
  unsigned short wSecond;
  unsigned short wMilliseconds;
}
SYSTEMTIME;

#define OUTPUT_HANDSET		0xc0
#define OUTPUT_HEADPHONE	0xc1
#define OUTPUT_SPEAKER		0xc2

#define VOLUME_LOW		0x01
#define VOLUME_LOW_SPEAKER	0x03
#define VOLUME_NORMAL		0x02
#define VOLUME_INSANELY_LOUD	0x07

#define MUTE_OFF		0x00
#define MUTE_ON			0xff
#define MUTE_ON_DISCRET		0xce

#define SIZE_HEADER 	6
#define SIZE_MAC_ADDR	17
#define TEXT_LENGTH_MAX	24
#define TEXT_LINE0		0x00
#define TEXT_LINE1		0x20
#define TEXT_LINE2		0x40
#define TEXT_NORMAL		0x05
#define TEXT_INVERSE	0x25

#define FAV_ICON_NONE	0x00
#define FAV_ICON_ONHOOK_BLACK		0x20
#define FAV_ICON_ONHOOK_WHITE		0x21
#define FAV_ICON_SPEAKER_ONHOOK_BLACK	0x22
#define FAV_ICON_SPEAKER_ONHOOK_WHITE	0x23
#define FAV_ICON_OFFHOOK_BLACK		0x24
#define FAV_ICON_OFFHOOK_WHITE		0x25
#define FAV_ICON_ONHOLD_BLACK		0x26
#define FAV_ICON_ONHOLD_WHITE		0x27
#define FAV_ICON_SPEAKER_OFFHOOK_BLACK	0x28
#define FAV_ICON_SPEAKER_OFFHOOK_WHITE	0x29
#define FAV_ICON_PHONE_BLACK		0x2A
#define FAV_ICON_PHONE_WHITE		0x2B
#define FAV_ICON_SPEAKER_ONHOLD_BLACK	0x2C
#define FAV_ICON_SPEAKER_ONHOLD_WHITE	0x2D
#define FAV_ICON_HEADPHONES		0x2E
#define FAV_ICON_HEADPHONES_ONHOLD	0x2F
#define FAV_ICON_HOME			0x30
#define FAV_ICON_CITY			0x31
#define FAV_ICON_SHARP			0x32
#define FAV_ICON_PAGER			0x33
#define FAV_ICON_CALL_CENTER		0x34
#define FAV_ICON_FAX			0x35
#define FAV_ICON_MAILBOX		0x36
#define FAV_ICON_REFLECT		0x37
#define FAV_ICON_COMPUTER		0x38
#define FAV_ICON_FORWARD		0x39
#define FAV_ICON_LOCKED			0x3A
#define FAV_ICON_TRASH			0x3B
#define FAV_ICON_INBOX			0x3C
#define FAV_ICON_OUTBOX			0x3D
#define FAV_ICON_MEETING		0x3E
#define FAV_ICON_BOX			0x3F

#define FAV_BLINK_FAST			0x20
#define FAV_BLINK_SLOW			0x40

#define FAV_MAX_LENGTH	0x0A
#define SIZE_PAGE 4096

// indent 132 col

/* Format of datagrams :
bytes 0 & 1 : ffff for discovery packet, 0000 for everything else
byte 2 : sequence number (high part)
byte 3 : sequence number (low part)
byte 4 : 2 = ask question or send info, 1 = answer or ACK, 0 = retransmit request
byte 5 : direction, 1 = server to phone, 2 = phone to server
arguments
*/

unsigned char PacketRcvDiscovery[] = { 0xff, 0xff, 0xff, 0xff, 0x02, 0x02, 0xff, 0xff, 0xff, 0xff, 0x9e, 0x03,
  0x08
};
unsigned char PacketSendDiscoveryACK[] = { 0x00, 0x00, /*Initial Seq (2 bytes) */ 0x00, 0x00, 0x00, 0x01 };

unsigned char PacketRecvFirmVersion[] = { 0x00, 0x00, 0x00, 0x13, 0x9a, 0x0a, 0x02 };
unsigned char PacketRecvPressedKey[] = { 0x00, 0x00, 0x00, 0x13, 0x99, 0x04, 0x00 };
unsigned char PacketRecvPickUp[] = { 0x00, 0x00, 0x00, 0x13, 0x99, 0x03, 0x04 };
unsigned char PacketRecvHangUp[] = { 0x00, 0x00, 0x00, 0x13, 0x99, 0x03, 0x03 };
unsigned char PacketRecvR2[] = { 0x00, 0x00, 0x00, 0x13, 0x96, 0x03, 0x03 };

// TransportAdapter
unsigned char PacketRecvResumeConnectionWithServer[] = { 0xff, 0xff, 0xff, 0xff, 0x9e, 0x03, 0x08 };
unsigned char PacketRecvMacAddr[] = { 0xff, 0xff, 0xff, 0xff, 0x9a, 0x0d, 0x07, 0x31, 0x38 /*MacAddr */  };

unsigned char PacketSendDateTime3[] = { 0x11, 0x09, 0x02, 0x02, /*Month */ 0x05, /*Day */ 0x06, /*Hour */ 0x07, /*Minutes */ 0x08, 0x32
};
unsigned char PacketSendDateTime[] = { 0x11, 0x09, 0x02, 0x0a, /*Month */ 0x05, /*Day */ 0x06, /*Hour */ 0x07,
/*Minutes */ 0x08, 0x32, 0x17, 0x04, 0x24, 0x07, 0x19, 0x04, 0x07, 0x00, 0x19, 0x05, 0x09, 0x3e, 0x0f,
  0x16, 0x05, 0x00, 0x80, 0x00, 0x1e, 0x05, 0x12, 0x00, 0x78
};

unsigned char PacketSendNoRing[] = { 0x16, 0x04, 0x1a, 0x00, 0x16, 0x04, 0x11, 0x00 };
unsigned char PacketSendS4[] = { 0x16, 0x04, 0x1a, 0x00, 0x16, 0x04, 0x11, 0x00, 0x16, 0x06, 0x32, 0xdf,
  0x00, 0xff, 0x16, 0x05, 0x1c, 0x00, 0x00, 0x17, 0x05, 0x0b, 0x00, 0x00, 0x19, 0x04, 0x00,
  0x00, 0x19, 0x04, 0x00, 0x08, 0x19, 0x04, 0x00, 0x10, 0x19, 0x04, 0x00, 0x18, 0x16, 0x05,
  0x31, 0x00, 0x00, 0x16, 0x05, 0x04, 0x00, 0x00
};
unsigned char PacketSendCall[] = { 0x16, 0x04, 0x1a, 0x00, 0x16, 0x04, 0x11, 0x00, 0x16, 0x06, 0x32, 0xdf,
  0x00, 0xff, 0x16, 0x05, 0x1c, 0x00, 0x00, 0x16, 0x0a, 0x38, 0x00, 0x12, 0xca, 0x03, 0xc0, 0xc3, 0xc5,
  0x16, 0x16, 0x30, 0x00, 0x00, /*codec */ 0x12, 0x12, /* frames per packet */ 0x01, 0x5c, 0x00, /*port RTP */ 0x0f,
  0xa0,
  /* port RTCP */ 0x9c, 0x41, /*port RTP */ 0x0f, 0xa0, /* port RTCP */ 0x9c, 0x41, /* IP Address */ 0x0a, 0x01, 0x16, 0x66
};
unsigned char PacketSendMute[] = { 0x16, 0x05, 0x04, 0x00, 0x00 };
unsigned char PacketSendStreamBasedToneOff[] = { 0x16, 0x05, 0x1c, 0x00, 0x00 };
unsigned char PacketSendCloseAudioStreamRX[] = { 0x16, 0x05, 0x31, 0x00, 0xff };
unsigned char PacketSendCloseAudioStreamTX[] = { 0x16, 0x05, 0x31, 0xff, 0x00 };
unsigned char PacketSendStreamBasedToneOn[] = { 0x16, 0x06, 0x1b, 0x00, 0x00, 0x05 };
unsigned char PacketSendStreamBasedToneSingleFreq[] = { 0x16, 0x06, 0x1d, 0x00, 0x01, 0xb8 };
unsigned char PacketSendStreamBasedToneDualFreq[] = { 0x16, 0x06, 0x1d, 0x00, 0x01, 0xb8, 0x01, 0x5e };
unsigned char PacketSendSelectOutput[] = { 0x16, 0x06, 0x32, 0xc0, 0x01, 0x00 };
unsigned char PacketSendEnableRecvr[] = { 0x16, 0x06, 0x32, 0xc0, 0x01, 0x00 };
unsigned char PacketSendDisableSpker[] = { 0x16, 0x06, 0x32, 0xc0, 0x03, 0x00 };
unsigned char PacketSendEnableSpker[] = { 0x16, 0x06, 0x32, 0xc2, 0x03, 0x00 };
unsigned char PacketSendRing[] = { 0x16, 0x06, 0x32, 0xdf, 0x00, 0xff, 0x16, 0x05, 0x1c, 0x00, 0x00, 0x16,
  0x04, 0x1a, 0x01, 0x16, 0x05, 0x12, 0x13 /* Ring type 10 to 17 */ , 0x18, 0x16, 0x04, 0x18,	/* volume 00, 10, 20... */
  0x20, 0x16, 0x04, 0x10, 0x00
};
unsigned char PacketSendEndCall[] = { 0x16, 0x06, 0x32, 0xdf, 0x00, 0xff, 0x16, 0x05, 0x31, 0x00, 0x00, 0x19,
  0x04, 0x00, 0x10, 0x19, 0x04, 0x00, 0x18, 0x16, 0x05, 0x04, 0x00, 0x00, 0x16, 0x04, 0x37, 0x10
};
unsigned char PacketSendS9[] = { 0x16, 0x06, 0x32, 0xdf, 0x00, 0xff, 0x19, 0x04, 0x00, 0x10, 0x16, 0x05, 0x1c, 0x00, 0x00 };
unsigned char PacketSendRTPpacketSize[] = { 0x16, 0x08, 0x38, 0x00, 0x00, 0xe0, 0x00, 0xa0 };
unsigned char PacketSendJitterBufferConf[] =
  { 0x16, 0x0e, 0x3a, 0x00, /* jitter */ 0x02, /* high water mark */ 0x04, 0x00, 0x00, /* early packet resync 2 bytes */ 0x3e, 0x80,
  0x00, 0x00, /* late packet resync 2 bytes */ 0x3e, 0x80
};

/* Duration in ms div 2 (0x20 = 64ms, 0x08 = 16ms) */
unsigned char PacketSendStreamBasedToneCad[] =
  { 0x16, 0x0a, 0x1e, 0x00, /* duration on */ 0x0a, /* duration off */ 0x0d, /* duration on */ 0x0a, /* duration off */ 0x0d,
  /* duration on */ 0x0a, /* duration off */ 0x2b
};
unsigned char PacketSendOpenAudioStreamRX[] =
  { 0x16, 0x1a, 0x30, 0x00, 0xff, /* Codec */ 0x00, 0x00, 0x01, 0x00, 0xb8, 0xb8, 0x0e, 0x0e, 0x01, /* Port */ 0x14,
  0x50, 0x00, 0x00, /* Port */ 0x14, 0x50, 0x00, 0x00, /* Dest IP */ 0x0a, 0x93, 0x69, 0x05
};
unsigned char PacketSendOpenAudioStreamTX[] =
  { 0x16, 0x1a, 0x30, 0xff, 0x00, 0x00, /* Codec */ 0x00, 0x01, 0x00, 0xb8, 0xb8, 0x0e, 0x0e, 0x01, /* Local port */ 0x14, 0x50,
  0x00, 0x00, /* Rmt Port */ 0x14, 0x50, 0x00, 0x00, /* Dest IP */ 0x0a, 0x93, 0x69, 0x05
};

unsigned char PacketSendArrow[] = { 0x17, 0x04, 0x04, 0x00 };
unsigned char PacketSendBlinkCursor[] = { 0x17, 0x04, 0x10, 0x86 };
unsigned char PacketSendDateTime2[] =
  { 0x17, 0x04, 0x17, 0x3d, 0x11, 0x09, 0x02, 0x0a, /*Month */ 0x05, /*Day */ 0x06, /*Hour */ 0x07, /*Minutes */ 0x08, 0x32 };
unsigned char PacketSendContrast[] = { 0x17, 0x04, 0x24, /*Contrast */ 0x08 };
unsigned char PacketSendStartTimer[] = { 0x17, 0x05, 0x0b, 0x05, 0x00, 0x17, 0x08, 0x16, 0x44, 0x75, 0x72, 0xe9, 0x65 };
unsigned char PacketSendStopTimer[] = { 0x17, 0x05, 0x0b, 0x02, 0x00 };
unsigned char PacketSendIcon[] = { 0x17, 0x05, 0x14, /*pos */ 0x00, /*icon */ 0x25 };	// display an icon in front of the text zone
unsigned char PacketSendS7[] = { 0x17, 0x06, 0x0f, 0x30, 0x07, 0x07 };
unsigned char PacketSendSetPosCursor[] = { 0x17, 0x06, 0x10, 0x81, 0x04, /*pos */ 0x20 };
unsigned char PacketSendMonthLabelsDownload[] = { 0x17, 0x0a, 0x15, /* Month (3 char) */ 0x46, 0x65, 0x62, 0x4d, 0xe4, 0x72, 0x20 };
unsigned char PacketSendFavorite[] = { 0x17, 0x0f, 0x19, 0x10, /*pos */ 0x01, /*name */ 0x20, 0x20, 0x20, 0x20,
  0x20, 0x20, 0x20, 0x20, 0x20, 0x20, /*end_name */ 0x19, 0x05, 0x0f,	/*pos */
  0x01, /*icone */ 0x00
};
unsigned char PacketSendTitle[] = { 0x17, 0x10, 0x19, 0x02, /*text */ 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
  0x20, 0x20, 0x20, 0x20, 0x20, 0x20	/*end_text */
};
unsigned char PacketSendText[] = { 0x17, 0x1e, 0x1b, 0x04, /*pos */ 0x00, /*inverse */ 0x25, /*text */ 0x20,
  0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
  0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, /*end_text */ 0x17, 0x04, 0x10, 0x87
};
unsigned char PacketSendStatus[] = { 0x17, 0x20, 0x19, 0x08, /*text */ 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
  0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
  0x20, 0x20, 0x20, 0x20	/*end_text */
};

unsigned char PacketSendLedUpdate[] = { 0x19, 0x04, 0x00, 0x00 };

unsigned char PacketSendQueryBasicManager04[] = { 0x1a, 0x04, 0x01, 0x04 };
unsigned char PacketSendQueryMacAddress[] = { 0x1a, 0x04, 0x01, 0x08 };
unsigned char PacketSendQueryBasicManager10[] = { 0x1a, 0x04, 0x01, 0x10 };
unsigned char PacketSendS1[] = { 0x1a, 0x07, 0x07, 0x00, 0x00, 0x00, 0x13 };

unsigned char PacketSendPing[] = { 0x1e, 0x05, 0x12, 0x00, /*Watchdog timer */ 0x78 };

unsigned char buffsend[64] = { 0x00, 0x00, 0xaa, 0xbb, 0x02, 0x01 };

struct tone_zone_unistim
{
  char country[3];
  int freq1;
  int freq2;
};

struct tone_zone_unistim frequency[] = {
  {"us", 350, 440},
  {"fr", 440, 0},
  {"au", 413, 438},
  {"nl", 425, 0},
  {"uk", 350, 440},
  {"fi", 425, 0},
  {"es", 425, 0},
  {"jp", 400, 0},
  {"no", 425, 0},
  {"at", 420, 0},
  {"nz", 400, 0},
  {"tw", 350, 440},
  {"cl", 400, 0},
  {"se", 425, 0},
  {"be", 425, 0},
  {"sg", 425, 0},
  {"il", 414, 0},
  {"br", 425, 0},
  {"hu", 425, 0},
  {"lt", 425, 0},
  {"pl", 425, 0},
  {"za", 400, 0},
  {"pt", 425, 0},
  {"ee", 425, 0},
  {"mx", 425, 0},
  {"in", 400, 0},
  {"de", 425, 0},
  {"ch", 425, 0},
  {"dk", 425, 0},
  {"cn", 450, 0},
  {"--", 0, 0}
};

// Global unistim */
struct sockaddr_in addr_from;	/* Global : give the IP address for the last packet received */
int size_addr_from = sizeof (addr_from);	/* Give the size of the sockaddr_in (used by WSARecvFrom) */
char *buff;			/* Receive buffer address */
unsigned char flag = 0;		/* Used for test */

struct unistim_subchannel
{
  ast_mutex_t lock;
  unsigned int callid;
  struct ast_channel *owner;
  struct unistim_line *parent;
  struct ast_rtp *rtp;
  time_t lastouttime;
  int progress;
  int ringing;
  int lastout;
  int nat;
  char ringvolume;
  char ringstyle;
  int outgoing;
  int alreadygone;
  struct unistim_subchannel *next;
};

struct unistim_line
{
  ast_mutex_t lock;
  char name[80];
  char label[11];		/* Label that shows next to the line buttons */
  struct unistim_subchannel *sub;	/* pointer to our current connection, channel and stuff */
  char accountcode[80];
  char exten[AST_MAX_EXTENSION];	/* Extention where to start */
  char context[AST_MAX_EXTENSION];
  char language[MAX_LANGUAGE];
  char cid_num[AST_MAX_EXTENSION];	/* Caller*ID */
  char cid_name[AST_MAX_EXTENSION];
  char lastcallerid[AST_MAX_EXTENSION];	/* Last Caller*ID */
  char call_forward[AST_MAX_EXTENSION];
  char mailbox[AST_MAX_EXTENSION];
  int lastmsgssent;
  time_t lastmsgcheck;
  char musicclass[MAX_MUSICCLASS];
  unsigned int callgroup;
  unsigned int pickupgroup;
  int callwaiting;
  int transfer;
  int threewaycalling;
  int cancallforward;
  int callreturn;
  int dnd;
  int hascallerid;
  int hidecallerid;
  int amaflags;
  int instance;
  int group;
  int needdestroy;
  int capability;
  int nonCodecCapability;
  int onhooktime;
  int msgstate;			/* voicemail message state */
  int immediate;
  int progress;
  struct unistim_line *next;
  struct unistim_device *parent;
};

static struct unistim_device
{
  /* A device containing one or more lines */
  int receiver_state;		// state of the receiver (see ReceiverState)
  int size_phone_number;	// size of the phone number
  char phone_number[16];	// the phone number entered by the user
  char redial_number[16];	// the last phone number entered by the user
  int phone_current;		// Number of the current phone
  int pos_fav;			// Position of the displayed favorites (used for scrolling)
  char id[18];			// mac address of the current phone in ascii
  char name[16];		// name of the device
  char softkeylabel[6][11];	// soft key label
  char softkeynumber[6][16];	// number dialed when the soft key is pressed
  char softkeyicon[6];		// icon number
  char softkeydevice[6][16];	// name of the device monitored
  struct unistim_device *sp[6];	// pointer to the device monitored by this soft key
  char maintext0[25];		// when the phone is idle, display this string on line 0
  char maintext1[25];		// when the phone is idle, display this string on line 1
  char maintext2[25];		// when the phone is idle, display this string on line 2
  char titledefault[13];	// title (text before date/time)
  char datetimeformat;		// format used for displaying time/date
  char contrast;		// contrast
  char country[3];		// country used for dial tone frequency
  char ringvolume;		// Ring volume
  char ringstyle;		// Ring melody
  int rtp_port;			// RTP port used by the phone
  int rtp_method;		// Select the unistim packet used for establishing a RTP session
  char codec_number;		// The current codec used to make calls
  int missed_call;		// Number of call unanswered
  int output;			// Handset, headphone or speaker
  int previous_output;		// Previous output
  int volume;			// Default volume
  int mute;			// Mute mode
  int moh;			// Music on hold in progress
  int size_buff_entry;		// size of the buffer used to enter datas
  char buff_entry[16];		// Buffer for temporary datas
  time_t start_call_timestamp;	// timestamp for the length calculation of the call
  struct unistim_line *lines;
  struct ast_ha *ha;
  struct unistimsession *session;
  struct unistim_device *next;
}
 *devices = NULL;

static struct unistimsession
{
  ast_mutex_t lock;
  struct sockaddr_in sin;	// IP address of the phone
  struct sockaddr_in sout;	// IP address of server
  int timeout;			// time-out in ticks : resend packet if no ack was received before the timeout occured
  unsigned short seq_phone;	// sequence number for the next packet (when we receive a request)
  unsigned short seq_server;	// sequence number for the next packet (when we send a request)
  unsigned short last_seq_ack;	// sequence number of the last ACK received
  unsigned long tick_next_ping;	// time for the next ping
  int last_buf_available;	// number of a free slot
  int nb_retransmit;		// number of retransmition
  int state;			// state of the phone (see PhoneState)
  char macaddr[18];		// mac adress of the phone (not always available)
  WSABUF wsabufsend[MAX_BUF_NUMBER];	// Size of each paquet stored in the buffer array & pointer to this buffer
  char buf[MAX_BUF_NUMBER][MAX_BUF_SIZE];	// Buffer array used to keep the lastest non-acked paquets
  struct unistim_device *device;
  struct unistimsession *next;
}
 *sessions = NULL;

// Proto
static struct ast_channel *unistim_new (struct unistim_subchannel *sub, int state);
static void ShowMainPage (struct unistimsession *pte);
static void SendRetransmit (struct unistimsession *pte);
static struct ast_channel *unistim_request (const char *type, int format, void *data, int *cause);
static int unistim_call (struct ast_channel *ast, char *dest, int timeout);
static int unistim_hangup (struct ast_channel *ast);
static int unistim_answer (struct ast_channel *ast);
static struct ast_frame *unistim_read (struct ast_channel *ast);
static int unistim_write (struct ast_channel *ast, struct ast_frame *frame);
static int unistim_indicate (struct ast_channel *ast, int ind);
static int unistim_fixup (struct ast_channel *oldchan, struct ast_channel *newchan);
static int unistim_senddigit (struct ast_channel *ast, char digit);
static int unistim_sendtext (struct ast_channel *ast, const char *text);

static const struct ast_channel_tech unistim_tech = {
  .type = type,
  .description = tdesc,
  .capabilities = CAPABILITY,
  .properties = AST_CHAN_TP_WANTSJITTER,
  .requester = unistim_request,
  .call = unistim_call,
  .hangup = unistim_hangup,
  .answer = unistim_answer,
  .read = unistim_read,
  .write = unistim_write,
  .indicate = unistim_indicate,
  .fixup = unistim_fixup,
  .send_digit = unistim_senddigit,
  .send_text = unistim_sendtext,
/*      .bridge = ast_rtp_bridge, */
};

static void
DisplayLastError (char *szMessage)
{
  int dwLastError;
  char *lpMsgBuf;
  time_t cur_time = time (0);

  dwLastError = errno;
  lpMsgBuf = strerror (dwLastError);

  // Display the error message
  ast_log (LOG_WARNING, "%s %s : (%u) %s\n", ctime (&cur_time), szMessage, dwLastError, lpMsgBuf);
}

static void
GetLocalTime (SYSTEMTIME * systime)
{
  struct tm *stm;
  time_t cur_time = time (0);

  if ((stm = localtime (&cur_time)) == 0)
    {
      DisplayLastError ("Error in localtime()");
      return;
    }
  systime->wYear = stm->tm_year;
  systime->wMonth = stm->tm_mon + 1;
  systime->wDayOfWeek = stm->tm_wday;
  systime->wDay = stm->tm_mday;
  systime->wHour = stm->tm_hour;
  systime->wMinute = stm->tm_min;
  systime->wSecond = stm->tm_sec;
  systime->wMilliseconds = 0;
  return;
}

static unsigned int
GetTickCount ()
{
  struct timeval tv;

  if (gettimeofday (&tv, NULL) == -1)
    DisplayLastError ("Error in gettimeofday");
  return (tv.tv_sec * 1000) + (tv.tv_usec / 1000);
}

// Send data to a phone without retransmit nor buffering
static void
SendRawClient (int size, char *data, struct sockaddr_in *addr_to, struct sockaddr_in *addr_ourip)
{
  struct iovec msg_iov;
  struct msghdr msg;
  char buffer[CMSG_SPACE (sizeof (struct in_pktinfo))];
  struct cmsghdr *ip_msg = (struct cmsghdr *) buffer;
  struct in_pktinfo *pki = (struct in_pktinfo *) CMSG_DATA (ip_msg);

  msg_iov.iov_base = data;
  msg_iov.iov_len = size;

  msg.msg_name = addr_to;	/* optional address */
  msg.msg_namelen = sizeof (struct sockaddr_in);	/* size of address */
  msg.msg_iov = &msg_iov;	/* scatter/gather array */
  msg.msg_iovlen = 1;		/* # elements in msg_iov */
  msg.msg_control = ip_msg;	/* ancillary data */
  msg.msg_controllen = sizeof (buffer);	/* ancillary data buffer len */
  msg.msg_flags = 0;		/* flags on received message */

  ip_msg->cmsg_len = CMSG_LEN (sizeof (*pki));
  ip_msg->cmsg_level = SOL_IP;
  ip_msg->cmsg_type = IP_PKTINFO;
  pki->ipi_ifindex = 0;		/* Interface index, 0 = use interface specified in routing table */
  pki->ipi_spec_dst.s_addr = addr_ourip->sin_addr.s_addr;	/* Local address */
  // pki->ipi_addr = ;  /* Header Destination address - ignored by kernel*/

#ifdef DUMP_PACKET
  if (unistimdebug)
    {
      int tmp;
      char iabuf[INET_ADDRSTRLEN];
      char iabuf2[INET_ADDRSTRLEN];
      ast_verbose ("\n**> From %s sending %d bytes to %s ***\n",
		   ast_inet_ntoa (iabuf2, sizeof (iabuf2), addr_ourip->sin_addr),
		   (int) size, ast_inet_ntoa (iabuf, sizeof (iabuf), addr_to->sin_addr));
      for (tmp = 0; tmp < size; tmp++)
	ast_verbose ("%.2x ", (unsigned char) data[tmp]);
      ast_verbose ("\n******************************************\n");

    }
#endif

  // if (sendto (unistimsock, data, size, 0, (struct sockaddr *) addr_to, sizeof (*addr_to)) == -1)
  if (sendmsg (unistimsock, &msg, 0) == -1)
    DisplayLastError ("Error sending datas");
}

static void
SendClient (int size, char *data, struct unistimsession *pte)
{
  struct unistimsession *cur = NULL;
  unsigned int tick;
  int buf_pos = pte->last_buf_available;
  unsigned short *sdata = (unsigned short *) data;

  if (buf_pos >= MAX_BUF_NUMBER)
    {
      ast_log (LOG_WARNING, "Error : send queue overflow\n");
      return;
    }
  sdata[1] = ntohs (++(pte->seq_server));
  pte->wsabufsend[buf_pos].len = size;
  memcpy (pte->wsabufsend[buf_pos].buf, data, size);

  tick = GetTickCount ();
  pte->timeout = tick + RETRANSMIT_TIMER;
  dwTimeOut = MAXINT;

  cur = sessions;
  while (cur)
    {
      if (dwTimeOut > cur->timeout - tick)
	dwTimeOut = cur->timeout - tick;
      cur = cur->next;
    }
#ifdef DUMP_PACKET
  if (unistimdebug)
    ast_verbose ("Using slot #%d :\n", buf_pos);
#endif

  SendRawClient (pte->wsabufsend[buf_pos].len, pte->wsabufsend[buf_pos].buf, &(pte->sin), &(pte->sout));
  pte->last_buf_available++;
}

static void
SendPing (struct unistimsession *pte)
{
  if ((unistimdebug) && (option_verbose > 5))
    ast_verbose ("Sending ping\n");
  pte->tick_next_ping = GetTickCount () + unistim_keepalive;
  memcpy (buffsend + SIZE_HEADER, PacketSendPing, sizeof (PacketSendPing));
  SendClient (SIZE_HEADER + sizeof (PacketSendPing), buffsend, pte);
}

static int
getToAddress (int fd, struct sockaddr_in *toAddr)
{
  int err;
  struct msghdr msg;
  struct
  {
    struct cmsghdr cm;
    int len;
    struct in_addr address;
  } ip_msg;

  /* Zero out the structures before we use them */
  /* This sets several key values to NULL */
  memset (&msg, 0, sizeof (msg));
  memset (&ip_msg, 0, sizeof (ip_msg));

  /* Initialize the message structure */
  msg.msg_control = &ip_msg;
  msg.msg_controllen = sizeof (ip_msg);
  /* Get info about the incoming packet */
  err = recvmsg (fd, &msg, MSG_PEEK);
  if (err == -1)
    ast_log (LOG_WARNING, "recvmsg returned an error: %s\n", strerror (errno));
  memcpy (&toAddr->sin_addr, &ip_msg.address, sizeof (struct in_addr));
  return err;
}

// Allocate memory & initialize structures for a new phone
// addr_from : ip address of the phone
static struct unistimsession *
CreateClient (struct sockaddr_in *addr_from)
{
  int tmp;
  struct unistimsession *s;

  s = malloc (sizeof (struct unistimsession));
  if (!s)
    {
      ast_log (LOG_WARNING, "Failed to allocate Unistim session: %s\n", strerror (errno));
      return NULL;
    }
  memset (s, 0, sizeof (struct unistimsession));
  memcpy (&s->sin, addr_from, sizeof (struct sockaddr_in));
  getToAddress (unistimsock, &s->sout);
  if (unistimdebug)
    {
      char iabuf[INET_ADDRSTRLEN];
      char iabuf2[INET_ADDRSTRLEN];

      ast_verbose ("Creating a new entry for the phone from %s received via server ip %s\n",
		   ast_inet_ntoa (iabuf, sizeof (iabuf), addr_from->sin_addr),
		   ast_inet_ntoa (iabuf2, sizeof (iabuf2), s->sout.sin_addr));
    }
  ast_mutex_init (&s->lock);
  ast_mutex_lock (&sessionlock);
  s->next = sessions;
  sessions = s;

  s->timeout = GetTickCount () + RETRANSMIT_TIMER;
  s->seq_phone = (short) 0x0000;
  s->seq_server = (short) 0x0000;
  s->last_seq_ack = (short) 0x000;
  s->last_buf_available = 0;
  s->nb_retransmit = 0;
  s->state = STATE_INIT;
  s->tick_next_ping = GetTickCount () + unistim_keepalive;
  // Initialize WSABUF
  for (tmp = 0; tmp < MAX_BUF_NUMBER; tmp++)
    {
      s->wsabufsend[tmp].buf = s->buf[tmp];
    }
  ast_mutex_unlock (&sessionlock);
  return s;
}


static struct unistim_subchannel *
find_subchannel_by_line (struct unistim_line *l)
{
  /* Change this function if you want three way call and stuff like that */
  struct unistim_subchannel *sub = l->sub;
  return sub;
}

static void
SendEndCall (struct unistimsession *pte)
{
  if (unistimdebug)
    ast_verbose ("Sending end call\n");
  memcpy (buffsend + SIZE_HEADER, PacketSendEndCall, sizeof (PacketSendEndCall));
  SendClient (SIZE_HEADER + sizeof (PacketSendEndCall), buffsend, pte);
}

static int
CallBackTimeOut (void *data)
{
  unsigned int tick;
  int i = 0;
  struct unistimsession *cur = NULL;

  taskid = -1;
  tick = GetTickCount ();
  DEBUG_TIMER ("Time-Out callback at %u\n", tick);
  ast_mutex_lock (&sessionlock);
  cur = sessions;
  while (cur)
    {
      DEBUG_TIMER ("Timeout for phone #%d at %u\n", i, cur->timeout);
      if (cur->timeout <= tick)
	{
	  // If the queue is empty, send a ping
	  if (cur->last_buf_available == 0)
	    SendPing (cur);
	  else
	    SendRetransmit (cur);
	}
      cur = cur->next;
      i++;
    }				// for
  ast_mutex_unlock (&sessionlock);
  return 0;
}

static void
SetPingTimer (struct unistimsession *pte)
{
  unsigned int tick;
  struct unistimsession *cur = NULL;
  if ((unistimdebug) && (option_verbose > 5))
    ast_verbose ("Setting ping timer\n");

  tick = GetTickCount ();
  pte->timeout = pte->tick_next_ping;
  DEBUG_TIMER ("tick = %u next ping at %u tick\n", tick, pte->timeout);
  if (tick > pte->timeout)
    {
      if (unistimdebug)
	ast_verbose ("A ping should be sent %d s ago\n", tick - pte->timeout);
      pte->timeout = tick;
    }
  dwTimeOut = MAXINT;
  cur = sessions;
  while (cur)
    {
      if (dwTimeOut > cur->timeout - tick)
	dwTimeOut = cur->timeout - tick;
      cur = cur->next;
    }

  DEBUG_TIMER ("taskid was == %d\n", taskid);
  if (taskid != -1)
    ast_sched_del (sched, taskid);
  if (dwTimeOut != MAXINT)
    taskid = ast_sched_add (sched, dwTimeOut, CallBackTimeOut, 0);
  if (taskid < 0)
    ast_log (LOG_WARNING, "ast_sched_add failed\n");
  DEBUG_TIMER ("taskid is == %d with time out == %d\n", taskid, dwTimeOut);
  return;
}
static void
SendStartTimer (struct unistimsession *pte)
{
  if (unistimdebug)
    ast_verbose ("Sending start timer\n");
  memcpy (buffsend + SIZE_HEADER, PacketSendStartTimer, sizeof (PacketSendStartTimer));
  SendClient (SIZE_HEADER + sizeof (PacketSendStartTimer), buffsend, pte);
}

static void
SendStopTimer (struct unistimsession *pte)
{
  if (unistimdebug)
    ast_verbose ("Sending stop timer\n");
  memcpy (buffsend + SIZE_HEADER, PacketSendStopTimer, sizeof (PacketSendStopTimer));
  SendClient (SIZE_HEADER + sizeof (PacketSendStopTimer), buffsend, pte);
}

static void
SendIcon (unsigned char pos, unsigned char status, struct unistimsession *pte)
{
  if (unistimdebug)
    ast_verbose ("Sending Icon pos %d with status 0x%.2x\n", pos, status);
  memcpy (buffsend + SIZE_HEADER, PacketSendIcon, sizeof (PacketSendIcon));
  buffsend[9] = pos;
  buffsend[10] = status;
  SendClient (SIZE_HEADER + sizeof (PacketSendIcon), buffsend, pte);
}

static void
SendTone (struct unistimsession *pte, uint16_t tone1, uint16_t tone2)
{
  if (!tone1)
    {
      if (unistimdebug)
	ast_verbose ("Sending Stream Based Tone Off\n");
      memcpy (buffsend + SIZE_HEADER, PacketSendStreamBasedToneOff, sizeof (PacketSendStreamBasedToneOff));
      SendClient (SIZE_HEADER + sizeof (PacketSendStreamBasedToneOff), buffsend, pte);
      return;
    }
  if (unistimdebug)
    ast_verbose ("Sending Stream Based Tone Cadence Download\n");
  memcpy (buffsend + SIZE_HEADER, PacketSendStreamBasedToneCad, sizeof (PacketSendStreamBasedToneCad));
//  SendClient (SIZE_HEADER + sizeof (PacketSendStreamBasedToneCad), buffsend, pte);
  if (unistimdebug)
    ast_verbose ("Sending Stream Based Tone Frequency Component List Download\n");
  tone1 *= 8;
  if (!tone2)
    {
      memcpy (buffsend + SIZE_HEADER, PacketSendStreamBasedToneSingleFreq, sizeof (PacketSendStreamBasedToneSingleFreq));
      buffsend[10] = (tone1 & 0xff00) >> 8;
      buffsend[11] = (tone1 & 0x00ff);
      SendClient (SIZE_HEADER + sizeof (PacketSendStreamBasedToneSingleFreq), buffsend, pte);
    }
  else
    {
      tone2 *= 8;
      memcpy (buffsend + SIZE_HEADER, PacketSendStreamBasedToneDualFreq, sizeof (PacketSendStreamBasedToneDualFreq));
      buffsend[10] = (tone1 & 0xff00) >> 8;
      buffsend[11] = (tone1 & 0x00ff);
      buffsend[12] = (tone2 & 0xff00) >> 8;
      buffsend[13] = (tone2 & 0x00ff);
      SendClient (SIZE_HEADER + sizeof (PacketSendStreamBasedToneDualFreq), buffsend, pte);
    }

  if (unistimdebug)
    ast_verbose ("Sending Stream Based Tone On\n");
  memcpy (buffsend + SIZE_HEADER, PacketSendStreamBasedToneOn, sizeof (PacketSendStreamBasedToneOn));
  SendClient (SIZE_HEADER + sizeof (PacketSendStreamBasedToneOn), buffsend, pte);
}

/* Positions for favorites
 |--------------------|
 |  5            2    |
 |  4            1    |
 |  3            0    |
*/

// status (icons) : 00 = nothing, 2x/3x = see parser.h, 4x/5x = blink fast, 6x/7x = blink slow
static void
SendFavorite (unsigned char pos, unsigned char status, struct unistimsession *pte, char *text)
{
  int i;

  if (unistimdebug)
    ast_verbose ("Sending Favorite pos %d with status 0x%.2x\n", pos, status);
  memcpy (buffsend + SIZE_HEADER, PacketSendFavorite, sizeof (PacketSendFavorite));
  buffsend[10] = pos;
  buffsend[24] = pos;
  buffsend[25] = status;
  i = strlen (text);
  if (i > FAV_MAX_LENGTH)
    i = FAV_MAX_LENGTH;
  memcpy (buffsend + FAV_MAX_LENGTH + 1, text, i);
  SendClient (SIZE_HEADER + sizeof (PacketSendFavorite), buffsend, pte);
}

static void
RefreshAllFavorite (struct unistimsession *pte)
{
  int i = 0;

  if (unistimdebug)
    ast_verbose ("Refreshing all favorite\n");
  SendFavorite ((unsigned char) i, pte->device->softkeyicon[i], pte, pte->device->softkeylabel[i]);
  for (i = 1; i < 6; i++)
    {
      if (pte->device->softkeyicon[i] <= FAV_ICON_HEADPHONES_ONHOLD)
	SendFavorite ((unsigned char) i, pte->device->softkeyicon[i] + 1, pte, pte->device->softkeylabel[i]);
      else
	SendFavorite ((unsigned char) i, pte->device->softkeyicon[i], pte, pte->device->softkeylabel[i]);

    }
}

// Change the status for this phone (pte) and update for each phones where pte is bookmarked
// use FAV_ICON_*_BLACK constant in status parameters
static void
ChangeFavoriteIcon (struct unistimsession *pte, unsigned char status)
{
  struct unistim_device *d = devices;
  int i;
  // Update the current phone
  if (pte->state != STATE_CLEANING)
    SendFavorite (0, status, pte, pte->device->softkeylabel[0]);
  // Notify other phones if we're in their bookmark
  while (d)
    {
      for (i = 1; i < 6; i++)
	{
	  if (d->sp[i] == pte->device)	/* It's us ? */
	    {
	      if (d->softkeyicon[i] != status)	/* Avoid resending the same icon */
		{
		  d->softkeyicon[i] = status;
		  if (d->session)
		    SendFavorite (i, status + 1, d->session, d->softkeylabel[i]);
		}
	    }
	}
      d = d->next;
    }
}

// Free memory allocated for a phone
static void
CloseClient (struct unistimsession *s)
{
  struct unistim_subchannel *sub;
  struct unistimsession *cur, *prev = NULL;
  ast_mutex_lock (&sessionlock);
  cur = sessions;
  /* Looking for the session in the linked chain */
  while (cur)
    {
      if (cur == s)
	break;
      prev = cur;
      cur = cur->next;
    }
  if (cur)			/* Session found ? */
    {
      if (cur->device)		/* This session was registred ? */
	{
	  s->state = STATE_CLEANING;
	  if (unistimdebug)
	    ast_verbose ("CloseClient session %p device %p lines %p sub %p\n", s, s->device, s->device->lines,
			 s->device->lines->sub);
	  ChangeFavoriteIcon (s, FAV_ICON_NONE);
	  sub = find_subchannel_by_line (s->device->lines);
	  if (sub)
	    {
	      if (sub->owner)	/* Call in progress ? */
		{
		  if (unistimdebug)
		    ast_verbose ("Aborting call\n");
		  ast_queue_hangup (sub->owner);
		}
	    }
	  else
	    ast_log (LOG_WARNING, "Freeing a client with no subchannel !\n");
	  cur->device->session = NULL;
	}
      else
	{
	  if (unistimdebug)
	    ast_verbose ("Freeing an unregistered client\n");
	}
      if (prev)
	prev->next = cur->next;
      else
	sessions = cur->next;
      ast_mutex_destroy (&s->lock);
      free (s);
    }
  else
    ast_log (LOG_WARNING, "Trying to delete non-existant session %p?\n", s);
  ast_mutex_unlock (&sessionlock);
  return;
}

static void
SendRetransmit (struct unistimsession *pte)
{
  int i;

  if (++pte->nb_retransmit >= NB_MAX_RETRANSMIT)
    {
      if (unistimdebug)
	ast_verbose ("Too many retransmit - freeing client\n");
      CloseClient (pte);
      return;
    }
  pte->timeout = GetTickCount () + RETRANSMIT_TIMER;

  for (i = pte->last_buf_available - (pte->seq_server - pte->last_seq_ack); i < pte->last_buf_available; i++)
    {
      if (unistimdebug)
	ast_verbose ("Retransmit slot #%d\n", i);
      SendRawClient (pte->wsabufsend[i].len, pte->wsabufsend[i].buf, &pte->sin, &pte->sout);
    }
  return;
}

// inverse : TEXT_INVERSE : yes, TEXT_NORMAL  : no
static void
SendText (unsigned char pos, unsigned char inverse, struct unistimsession *pte, const char *text)
{
  int i;

  if (unistimdebug)
    ast_verbose ("Sending text at pos %d, inverse flag %d\n", pos, inverse);
  memcpy (buffsend + SIZE_HEADER, PacketSendText, sizeof (PacketSendText));
  buffsend[10] = pos;
  buffsend[11] = inverse;
  i = strlen (text);
  if (i > TEXT_LENGTH_MAX)
    i = TEXT_LENGTH_MAX;
  memcpy (buffsend + 12, text, i);
  SendClient (SIZE_HEADER + sizeof (PacketSendText), buffsend, pte);
}

static void
SendTextStatus (struct unistimsession *pte, char *text)
{
  int i;
  if (unistimdebug)
    ast_verbose ("Sending status text\n");
  memcpy (buffsend + SIZE_HEADER, PacketSendStatus, sizeof (PacketSendStatus));
  i = strlen (text);
  if (i > 28)
    i = 28;
  memcpy (buffsend + 10, text, i);
  SendClient (SIZE_HEADER + sizeof (PacketSendStatus), buffsend, pte);
}

/* led values in hexa : 0 = bar off, 1 = bar on, 2 = bar 1s on/1s off, 3 = bar 2.5s on/0.5s off
 * 4 = bar 0.6s on/0.3s off, 5 = bar 0.5s on/0.5s off, 6 = bar 2s on/0.5s off
 * 7 = bar off, 8 = speaker off, 9 = speaker on, 10 = headphone off, 11 = headphone on
 * 18 = mute off, 19 mute on */
static void
SendLedUpdate (struct unistimsession *pte, unsigned char led)
{
  if (unistimdebug)
    ast_verbose ("Sending LedUpdate (%x)\n", led);
  memcpy (buffsend + SIZE_HEADER, PacketSendLedUpdate, sizeof (PacketSendLedUpdate));
  buffsend[9] = led;
  SendClient (SIZE_HEADER + sizeof (PacketSendLedUpdate), buffsend, pte);
}

/* output = OUTPUT_HANDSET, OUTPUT_HEADPHONE or OUTPUT_SPEAKER
 * volume = VOLUME_LOW, VOLUME_NORMAL, VOLUME_INSANELY_LOUD
 * mute = MUTE_OFF, MUTE_ON */
static void
SendSelectOutput (struct unistimsession *pte, unsigned char output, unsigned char volume, unsigned char mute)
{
  if (unistimdebug)
    ast_verbose ("Sending select output packet output=%x volume=%x mute=%x\n", output, volume, mute);
  memcpy (buffsend + SIZE_HEADER, PacketSendSelectOutput, sizeof (PacketSendSelectOutput));
  buffsend[9] = output;
  if (output == OUTPUT_SPEAKER)
    volume = VOLUME_LOW_SPEAKER;
  else
    volume = VOLUME_LOW;
  buffsend[10] = volume;
  if (mute == MUTE_ON_DISCRET)
    buffsend[11] = MUTE_ON;
  else
    buffsend[11] = mute;
  SendClient (SIZE_HEADER + sizeof (PacketSendSelectOutput), buffsend, pte);
  if (mute == MUTE_OFF)
    SendLedUpdate (pte, 0x18);
  else if (mute == MUTE_ON)
    SendLedUpdate (pte, 0x19);
  pte->device->mute = mute;
  if (output == OUTPUT_HANDSET)
    {
      if (mute == MUTE_ON)
	ChangeFavoriteIcon (pte, FAV_ICON_ONHOLD_BLACK);
      else
	ChangeFavoriteIcon (pte, FAV_ICON_OFFHOOK_BLACK);
      SendLedUpdate (pte, 0x08);
      SendLedUpdate (pte, 0x10);
    }
  else if (output == OUTPUT_HEADPHONE)
    {
      if (mute == MUTE_ON)
	ChangeFavoriteIcon (pte, FAV_ICON_HEADPHONES_ONHOLD);
      else
	ChangeFavoriteIcon (pte, FAV_ICON_HEADPHONES);
      SendLedUpdate (pte, 0x08);
      SendLedUpdate (pte, 0x11);
    }
  else if (output == OUTPUT_SPEAKER)
    {
      SendLedUpdate (pte, 0x10);
      SendLedUpdate (pte, 0x09);
      if (pte->device->receiver_state == STATE_OFFHOOK)
	{
	  if (mute == MUTE_ON)
	    ChangeFavoriteIcon (pte, FAV_ICON_SPEAKER_ONHOLD_BLACK);
	  else
	    ChangeFavoriteIcon (pte, FAV_ICON_SPEAKER_ONHOOK_BLACK);
	}
      else
	{
	  if (mute == MUTE_ON)
	    ChangeFavoriteIcon (pte, FAV_ICON_SPEAKER_ONHOLD_BLACK);
	  else
	    ChangeFavoriteIcon (pte, FAV_ICON_SPEAKER_OFFHOOK_BLACK);
	}
    }
  else
    ast_log (LOG_WARNING, "Invalid ouput (%d)\n", output);
  if (output != pte->device->output)
    pte->device->previous_output = pte->device->output;
  pte->device->output = output;
}

static void
SendRing (struct unistimsession *pte, char volume, char style)
{
  if (unistimdebug)
    ast_verbose ("Sending ring packet\n");
  memcpy (buffsend + SIZE_HEADER, PacketSendRing, sizeof (PacketSendRing));
  buffsend[24] = style + 0x10;
  buffsend[29] = volume * 0x10;
  SendClient (SIZE_HEADER + sizeof (PacketSendRing), buffsend, pte);
}

static void
SendNoRing (struct unistimsession *pte)
{
  if (unistimdebug)
    ast_verbose ("Sending no ring packet\n");
  memcpy (buffsend + SIZE_HEADER, PacketSendNoRing, sizeof (PacketSendNoRing));
  SendClient (SIZE_HEADER + sizeof (PacketSendNoRing), buffsend, pte);
}

static void
SendTextTitle (struct unistimsession *pte, char *text)
{
  int i;
  if (unistimdebug)
    ast_verbose ("Sending title text\n");
  memcpy (buffsend + SIZE_HEADER, PacketSendTitle, sizeof (PacketSendTitle));
  i = strlen (text);
  if (i > 12)
    i = 12;
  memcpy (buffsend + 10, text, i);
  SendClient (SIZE_HEADER + sizeof (PacketSendTitle), buffsend, pte);

}

static void
SendTest (struct unistimsession *pte)
{
  char text[TEXT_LENGTH_MAX];
  int i;
  if (unistimdebug)
    ast_verbose ("Sending test packet %x\n", flag);

  sprintf (text, "first/last char0x%x/0x%x", flag, flag + TEXT_LENGTH_MAX - 1);
  SendText (TEXT_LINE1, TEXT_NORMAL, pte, text);

  for (i = 0; i < TEXT_LENGTH_MAX; i++)
    {
      text[i] = flag++;
    }
  text[i] = '\0';
  SendText (TEXT_LINE0, TEXT_NORMAL, pte, text);
}
static void
SendDateTime (struct unistimsession *pte)
{
  SYSTEMTIME systime;
  if (unistimdebug)
    ast_verbose ("Sending Time & Date\n");
  memcpy (buffsend + SIZE_HEADER, PacketSendDateTime, sizeof (PacketSendDateTime));
  GetLocalTime (&systime);
  buffsend[10] = (unsigned char) systime.wMonth;
  buffsend[11] = (unsigned char) systime.wDay;
  buffsend[12] = (unsigned char) systime.wHour;
  buffsend[13] = (unsigned char) systime.wMinute;
  SendClient (SIZE_HEADER + sizeof (PacketSendDateTime), buffsend, pte);
}

static void
SendDateTime2 (struct unistimsession *pte)
{
  SYSTEMTIME systime;
  if (unistimdebug)
    ast_verbose ("Sending Time & Date #2\n");
  memcpy (buffsend + SIZE_HEADER, PacketSendDateTime2, sizeof (PacketSendDateTime2));
  GetLocalTime (&systime);
  if (pte->device)
    buffsend[9] = pte->device->datetimeformat;
  else
    buffsend[9] = 61;
  buffsend[14] = (unsigned char) systime.wMonth;
  buffsend[15] = (unsigned char) systime.wDay;
  buffsend[16] = (unsigned char) systime.wHour;
  buffsend[17] = (unsigned char) systime.wMinute;
  SendClient (SIZE_HEADER + sizeof (PacketSendDateTime2), buffsend, pte);
}

static void
SendDateTime3 (struct unistimsession *pte)
{
  SYSTEMTIME systime;
  if (unistimdebug)
    ast_verbose ("Sending Time & Date #3\n");
  memcpy (buffsend + SIZE_HEADER, PacketSendDateTime3, sizeof (PacketSendDateTime3));
  GetLocalTime (&systime);
  buffsend[10] = (unsigned char) systime.wMonth;
  buffsend[11] = (unsigned char) systime.wDay;
  buffsend[12] = (unsigned char) systime.wHour;
  buffsend[13] = (unsigned char) systime.wMinute;
  SendClient (SIZE_HEADER + sizeof (PacketSendDateTime3), buffsend, pte);
}

static void
SendBlinkCursor (struct unistimsession *pte)
{
  if (unistimdebug)
    ast_verbose ("Sending set blink\n");
  memcpy (buffsend + SIZE_HEADER, PacketSendBlinkCursor, sizeof (PacketSendBlinkCursor));
  SendClient (SIZE_HEADER + sizeof (PacketSendBlinkCursor), buffsend, pte);
  return;
}

// pos : 0xab (a=0/2/4 = line ; b = row)
static void
SendCursorPos (struct unistimsession *pte, unsigned char pos)
{
  if (unistimdebug)
    ast_verbose ("Sending set cursor position\n");
  memcpy (buffsend + SIZE_HEADER, PacketSendSetPosCursor, sizeof (PacketSendSetPosCursor));
  buffsend[11] = pos;
  SendClient (SIZE_HEADER + sizeof (PacketSendSetPosCursor), buffsend, pte);
  return;
}

static void
RcvResumeConnectionWithServer (struct unistimsession *pte)
{
  if (unistimdebug)
    ast_verbose ("ResumeConnectionWithServer received\n");
  if (unistimdebug)
    ast_verbose ("Sending PacketSendQueryMacAddress\n");
  memcpy (buffsend + SIZE_HEADER, PacketSendQueryMacAddress, sizeof (PacketSendQueryMacAddress));
  SendClient (SIZE_HEADER + sizeof (PacketSendQueryMacAddress), buffsend, pte);
  return;
}


static int
unistim_register (char *buf, struct unistimsession *s)
{
  struct unistim_device *d;

  ast_mutex_lock (&devicelock);
  d = devices;
  while (d)
    {
      if (!strcasecmp (s->macaddr, d->id))
	{
	  /* XXX Deal with IP authentication */
	  s->device = d;
	  d->session = s;
	  d->codec_number = DEFAULT_CODEC;
	  d->pos_fav = 0;
	  d->missed_call = 0;
	  d->receiver_state = STATE_ONHOOK;

	  break;
	}
      d = d->next;
    }
  ast_mutex_unlock (&devicelock);

  if (!d)
    return 0;

  return 1;
}

static void
RcvMacAddr (struct unistimsession *pte, char *buf)
{
  int tmp, i = 0;
  char addrmac[19];
  int res = 0;
  if (unistimdebug)
    ast_verbose ("Mac Address received : ");
  for (tmp = 15; tmp < 15 + SIZE_HEADER; tmp++)
    {
      sprintf (&addrmac[i], "%.2x", (unsigned char) buf[tmp]);
      i += 2;
    }
  if (unistimdebug)
    ast_verbose ("%s\n", addrmac);
  strcpy (pte->macaddr, addrmac);
  res = unistim_register (buf, pte);
  if (!res)
    {
      ast_log (LOG_WARNING, "No entry found for this phone : %s\n", addrmac);
      pte->state = STATE_AUTHDENY;
    }
  else
    {
      pte->state = STATE_MAINPAGE;
      if (option_verbose > 2)
	ast_verbose (VERBOSE_PREFIX_3 "Device '%s' successfuly registered\n", pte->device->name);
    }

  if (unistimdebug)
    ast_verbose ("\nSending S1\n");
  memcpy (buffsend + SIZE_HEADER, PacketSendS1, sizeof (PacketSendS1));
  SendClient (SIZE_HEADER + sizeof (PacketSendS1), buffsend, pte);

  if (unistimdebug)
    ast_verbose ("Sending QueryBasicManager04\n");
  memcpy (buffsend + SIZE_HEADER, PacketSendQueryBasicManager04, sizeof (PacketSendQueryBasicManager04));
  SendClient (SIZE_HEADER + sizeof (PacketSendQueryBasicManager04), buffsend, pte);

  if (unistimdebug)
    ast_verbose ("Sending QueryBasicManager10\n");
  memcpy (buffsend + SIZE_HEADER, PacketSendQueryBasicManager10, sizeof (PacketSendQueryBasicManager10));
  SendClient (SIZE_HEADER + sizeof (PacketSendQueryBasicManager10), buffsend, pte);

  SendDateTime (pte);
  return;
}

static void
CancelDial (struct unistimsession *pte)
{
  SendNoRing (pte);
  pte->device->missed_call++;
  ShowMainPage (pte);
  return;
}

static void
CloseCall (struct unistimsession *pte)
{
  struct unistim_subchannel *sub;
  sub = find_subchannel_by_line (pte->device->lines);
  SendEndCall (pte);
  SendStopTimer (pte);
  if (sub->owner)
    {
      sub->alreadygone = 1;
      ast_queue_hangup (sub->owner);
    }
  else
    ast_log (LOG_WARNING, "USTM(%s@%s-%d) channel already destroyed\n", sub->parent->name, sub->parent->parent->name, sub->callid);
  ShowMainPage (pte);
  return;
}

static void
IgnoreCall (struct unistimsession *pte)
{
  SendNoRing (pte);
  return;
}

static void *
unistim_ss (void *data)
{
  struct ast_channel *chan = data;
  struct unistim_subchannel *sub = chan->tech_pvt;
  struct unistim_line *l = sub->parent;
  struct unistimsession *s = l->parent->session;
  int res;

  if (option_verbose > 2)
    ast_verbose (VERBOSE_PREFIX_3 "Starting simple switch on '%s@%s'\n", l->name, l->parent->name);
  strncpy (chan->exten, s->device->phone_number, sizeof (chan->exten) - 1);
  strncpy (s->device->redial_number, s->device->phone_number, sizeof (s->device->redial_number) - 1);
  if (!ast_strlen_zero (l->cid_num))
    {
      if (!l->hidecallerid)
	chan->cid.cid_num = strdup (l->cid_num);
      chan->cid.cid_ani = strdup (l->cid_num);
    }
  ast_setstate (chan, AST_STATE_RING);
  res = ast_pbx_run (chan);
  if (res)
    {
      ast_log (LOG_WARNING, "PBX exited non-zero\n");
      SendTone (s, 1000, 0);;
    }
  return NULL;
}

static void
start_rtp (struct unistim_subchannel *sub)
{
  struct sockaddr_in us;
  struct sockaddr_in public;
  struct sockaddr_in sin;
  char iabuf[INET_ADDRSTRLEN];
  int codec;
  struct sockaddr_in sout = sub->parent->parent->session->sout;

  ast_mutex_lock (&sub->lock);
  /* Allocate the RTP */
  if (unistimdebug)
    ast_verbose ("Starting RTP. Bind on %s\n", ast_inet_ntoa (iabuf, sizeof (iabuf), sout.sin_addr));
  sub->rtp = ast_rtp_new_with_bindaddr (sched, io, 1, 0, sout.sin_addr);
  if (!sub->rtp)
    {
      ast_log (LOG_WARNING, "Unable to create RTP session: %s binaddr=%s\n", strerror (errno),
	       ast_inet_ntoa (iabuf, sizeof (iabuf), sout.sin_addr));
      return;
    }
  if (sub->rtp && sub->owner)
    {
      sub->owner->fds[0] = ast_rtp_fd (sub->rtp);
      sub->owner->fds[1] = ast_rtcp_fd (sub->rtp);
    }
  if (sub->rtp)
    ast_rtp_setnat (sub->rtp, sub->nat);

  /* Create the RTP connection */
  ast_rtp_get_us (sub->rtp, &us);
  sin.sin_family = AF_INET;
  /* Setting up RTP for our side */
  memcpy (&sin.sin_addr, &sub->parent->parent->session->sin.sin_addr, sizeof (sin.sin_addr));
  sin.sin_port = htons (sub->parent->parent->rtp_port);
  ast_rtp_set_peer (sub->rtp, &sin);
  if (!(sub->owner->nativeformats & sub->owner->readformat))
    {
      int fmt;
      fmt = ast_best_codec (sub->owner->nativeformats);
      ast_log (LOG_WARNING,
	       "Our read/writeformat has been changed to something incompatible : %s (%d), using %s (%d) best codec from %d\n",
	       ast_getformatname (sub->owner->readformat), sub->owner->readformat, ast_getformatname (fmt), fmt,
	       sub->owner->nativeformats);
      sub->owner->readformat = fmt;
      sub->owner->writeformat = fmt;
    }
  codec = ast_rtp_lookup_code (sub->rtp, 1, sub->owner->readformat);
  /* Setting up RTP of the phone */
  if (public_ip.sin_family == 0)	// NAT IP override ?
    memcpy (&public, &us, sizeof (public));	// No defined, using IP from recvmsg
  else
    memcpy (&public, &public_ip, sizeof (public));	// override
  if (unistimdebug)
    {
      ast_verbose ("RTP started : Our IP/port is : %s:%hd with codec %s (%d)\n",
		   ast_inet_ntoa (iabuf, sizeof (iabuf), us.sin_addr), htons (us.sin_port),
		   ast_getformatname (sub->owner->readformat), sub->owner->readformat);
      ast_verbose ("Starting phone RTP stack. Our public IP is %s\n", ast_inet_ntoa (iabuf, sizeof (iabuf), public.sin_addr));
    }
  if ((sub->owner->readformat == AST_FORMAT_ULAW) || (sub->owner->readformat == AST_FORMAT_ALAW))
    {
      if (unistimdebug)
	ast_verbose ("Sending PacketSendRTPpacketSize for codec %d\n", codec);
      memcpy (buffsend + SIZE_HEADER, PacketSendRTPpacketSize, sizeof (PacketSendRTPpacketSize));
      buffsend[10] = codec;
      SendClient (SIZE_HEADER + sizeof (PacketSendRTPpacketSize), buffsend, sub->parent->parent->session);
    }
  if (unistimdebug)
    ast_verbose ("Sending Jitter Buffer Parameters Configuration\n");
  memcpy (buffsend + SIZE_HEADER, PacketSendJitterBufferConf, sizeof (PacketSendJitterBufferConf));
  SendClient (SIZE_HEADER + sizeof (PacketSendJitterBufferConf), buffsend, sub->parent->parent->session);
  if (sub->parent->parent->rtp_method != 0)
    {
      if (unistimdebug)
	ast_verbose ("Sending OpenAudioStreamTX using method #%d\n", sub->parent->parent->rtp_method);
      memcpy (buffsend + SIZE_HEADER, PacketSendOpenAudioStreamTX, sizeof (PacketSendOpenAudioStreamTX));
      if (sub->parent->parent->rtp_method == 1)
	{
	  memcpy (buffsend + 28, &public.sin_addr, sizeof (public.sin_addr));
	  buffsend[20] = (htons (sin.sin_port) & 0xff00) >> 8;
	  buffsend[21] = (htons (sin.sin_port) & 0x00ff);
	  buffsend[25] = (us.sin_port & 0xff00) >> 8;
	  buffsend[24] = (us.sin_port & 0x00ff);
	}
      else
	{
	  memcpy (buffsend + 23, &public.sin_addr, sizeof (public.sin_addr));
	  buffsend[15] = (htons (sin.sin_port) & 0xff00) >> 8;
	  buffsend[16] = (htons (sin.sin_port) & 0x00ff);
	  buffsend[20] = (us.sin_port & 0xff00) >> 8;
	  buffsend[19] = (us.sin_port & 0x00ff);
	}
      SendClient (SIZE_HEADER + sizeof (PacketSendOpenAudioStreamTX), buffsend, sub->parent->parent->session);

      if (unistimdebug)
	ast_verbose ("Sending OpenAudioStreamRX\n");
      memcpy (buffsend + SIZE_HEADER, PacketSendOpenAudioStreamRX, sizeof (PacketSendOpenAudioStreamRX));
      if (sub->parent->parent->rtp_method == 1)
	{
	  memcpy (buffsend + 28, &public.sin_addr, sizeof (public.sin_addr));
	  buffsend[20] = (htons (sin.sin_port) & 0xff00) >> 8;
	  buffsend[21] = (htons (sin.sin_port) & 0x00ff);
	  buffsend[25] = (us.sin_port & 0xff00) >> 8;
	  buffsend[24] = (us.sin_port & 0x00ff);
	}
      else
	{
	  memcpy (buffsend + 23, &public.sin_addr, sizeof (public.sin_addr));
	  buffsend[15] = (htons (sin.sin_port) & 0xff00) >> 8;
	  buffsend[16] = (htons (sin.sin_port) & 0x00ff);
	  buffsend[20] = (us.sin_port & 0xff00) >> 8;
	  buffsend[19] = (us.sin_port & 0x00ff);
	}
      SendClient (SIZE_HEADER + sizeof (PacketSendOpenAudioStreamRX), buffsend, sub->parent->parent->session);
    }
  else
    {
      uint16_t rtcpsin_port = htons (us.sin_port) + 1;	// RTCP port is RTP + 1

      if (unistimdebug)
	ast_verbose ("Sending PacketSendCall default method\n");

      memcpy (buffsend + SIZE_HEADER, PacketSendCall, sizeof (PacketSendCall));
      memcpy (buffsend + 53, &public.sin_addr, sizeof (public.sin_addr));
      /* Destination port when sending RTP */
      buffsend[49] = (us.sin_port & 0x00ff);
      buffsend[50] = (us.sin_port & 0xff00) >> 8;
      /* Destination port when sending RTCP */
      buffsend[52] = (rtcpsin_port & 0x00ff);
      buffsend[51] = (rtcpsin_port & 0xff00) >> 8;
      /* Codec */
      buffsend[40] = codec;
      buffsend[41] = codec;
      if (sub->owner->readformat == AST_FORMAT_ULAW)
	buffsend[42] = 1;	// 1 = 20ms (160 bytes), 2 = 40ms (320 bytes)
      else if (sub->owner->readformat == AST_FORMAT_ALAW)
	buffsend[42] = 1;	// 1 = 20ms (160 bytes), 2 = 40ms (320 bytes)
      else if (sub->owner->readformat == AST_FORMAT_G723_1)
	buffsend[42] = 2;	// 1 = 30ms (24 bytes), 2 = 60 ms (48 bytes)
      else if (sub->owner->readformat == AST_FORMAT_G729A)
	buffsend[42] = 2;	// 1 = 10ms (10 bytes), 2 = 20ms (20 bytes)
      else
	ast_log (LOG_WARNING, "Unsupported codec %s (%d) !\n", ast_getformatname (sub->owner->readformat), sub->owner->readformat);
      /* Source port for transmit RTP and Destination port for receiving RTP */
      buffsend[45] = (htons (sin.sin_port) & 0xff00) >> 8;
      buffsend[46] = (htons (sin.sin_port) & 0x00ff);
      buffsend[47] = (rtcpsin_port & 0xff00) >> 8;
      buffsend[48] = (rtcpsin_port & 0x00ff);
      SendClient (SIZE_HEADER + sizeof (PacketSendCall), buffsend, sub->parent->parent->session);
    }
  ast_mutex_unlock (&sub->lock);
}

/* From phone to PBX */
static void
HandleCallOutgoing (struct unistimsession *s)
{
  struct ast_channel *c;
  struct unistim_subchannel *sub;
  pthread_t t;
  s->state = STATE_CALL;
  sub = find_subchannel_by_line (s->device->lines);
  if (!sub)
    {
      ast_log (LOG_NOTICE, "No available lines on: %s\n", s->device->name);
      return;
    }
  if (!sub->owner)
    {
      c = unistim_new (sub, AST_STATE_DOWN);
      if (c)
	{
	  /* Need to start RTP before calling ast_pbx_run */
	  if (!sub->rtp)
	    start_rtp (sub);
	  SendSelectOutput (s, s->device->output, s->device->volume, MUTE_OFF);
	  SendText (TEXT_LINE0, TEXT_NORMAL, s, "Calling :");
	  SendText (TEXT_LINE1, TEXT_NORMAL, s, s->device->phone_number);
	  SendText (TEXT_LINE2, TEXT_NORMAL, s, "Dialing...");
	  SendTextStatus (s, "Hangup");
	  /* start switch */
	  if (ast_pthread_create (&t, NULL, unistim_ss, c))
	    {
	      DisplayLastError ("Unable to create switch thread");
	      ast_queue_hangup (c);
	    }
	}
      else
	ast_log (LOG_WARNING, "Unable to create channel for %s@%s\n", sub->parent->name, s->device->name);
    }
  else
    ast_log (LOG_DEBUG, "Current sub [%s] already has owner\n", sub->owner->name);
  return;
}

/* From PBX to phone */
static void
HandleCallIncoming (struct unistimsession *s)
{
  struct unistim_subchannel *sub;
  s->state = STATE_CALL;
  s->device->missed_call = 0;
  SendNoRing (s);
  sub = find_subchannel_by_line (s->device->lines);
  if (!sub)
    {
      ast_log (LOG_NOTICE, "No available lines on: %s\n", s->device->name);
      return;
    }
  sub->outgoing = 1;
  start_rtp (sub);
  if (sub->outgoing)
    ast_queue_control (sub->owner, AST_CONTROL_ANSWER);
  else
    ast_log (LOG_WARNING, "Unable to create channel for %s@%s\n", sub->parent->name, s->device->name);
  SendText (TEXT_LINE2, TEXT_NORMAL, s, "is on-line");
  SendTextStatus (s, "Hangup");
  SendStartTimer (s);

  if ((s->device->output == OUTPUT_HANDSET) && (s->device->receiver_state == STATE_ONHOOK))
    SendSelectOutput (s, OUTPUT_SPEAKER, s->device->volume, MUTE_OFF);
  else
    SendSelectOutput (s, s->device->output, s->device->volume, MUTE_OFF);
  s->device->start_call_timestamp = time (0);
  return;
}

static void
KeyCall (struct unistimsession *pte, char keycode)
{
  struct ast_frame f = { 0, };
  if ((keycode >= KEY_0) && (keycode <= KEY_SHARP))
    {
      struct unistim_subchannel *sub;
      sub = find_subchannel_by_line (pte->device->lines);
      if (!sub->owner)
	{
	  ast_log (LOG_WARNING, "Unable to find subchannel in dtmf senddigit\n");
	  return;
	}
      if (keycode == KEY_SHARP)
	keycode = '#';
      else if (keycode == KEY_STAR)
	keycode = '*';
      else
	keycode -= 0x10;
      if (unistimdebug)
	ast_verbose ("Send Digit %c", keycode);
      ast_senddigit (sub->owner, keycode);
      f.frametype = AST_FRAME_DTMF;
      f.subclass = keycode;
      f.src = "unistim";
      ast_queue_frame (sub->owner, &f);
      return;
    }
  switch (keycode)
    {
    case KEY_HANGUP:
    case KEY_FUNC1:
      CloseCall (pte);
      break;
    case KEY_HEADPHN:
      if (pte->device->output == OUTPUT_HEADPHONE)
	SendSelectOutput (pte, OUTPUT_HANDSET, pte->device->volume, MUTE_OFF);
      else
	SendSelectOutput (pte, OUTPUT_HEADPHONE, pte->device->volume, MUTE_OFF);
      break;
    case KEY_LOUDSPK:
      if (pte->device->output != OUTPUT_SPEAKER)
	SendSelectOutput (pte, OUTPUT_SPEAKER, pte->device->volume, MUTE_OFF);
      else
	SendSelectOutput (pte, pte->device->previous_output, pte->device->volume, MUTE_OFF);
      break;
    case KEY_MUTE:
      if (!pte->device->moh)
	{
	  if (pte->device->mute == MUTE_ON)
	    SendSelectOutput (pte, pte->device->output, pte->device->volume, MUTE_OFF);
	  else
	    SendSelectOutput (pte, pte->device->output, pte->device->volume, MUTE_ON);
	  break;
	}
    case KEY_ONHOLD:
      {
	struct unistim_subchannel *sub;
	struct ast_channel *bridgepeer = NULL;
	sub = find_subchannel_by_line (pte->device->lines);
	if (!sub->owner)
	  {
	    ast_log (LOG_WARNING, "Unable to find subchannel for music on hold\n");
	    return;
	  }
	if ((bridgepeer = ast_bridged_channel (sub->owner)))
	  {
	    if (pte->device->moh)
	      {
		ast_moh_stop (bridgepeer);
		pte->device->moh = 0;
		SendSelectOutput (pte, pte->device->output, pte->device->volume, MUTE_OFF);
	      }
	    else
	      {
		ast_moh_start (bridgepeer, pte->device->lines->musicclass);
		pte->device->moh = 1;
		SendSelectOutput (pte, pte->device->output, pte->device->volume, MUTE_ON);
	      }
	  }
	else
	  ast_log (LOG_WARNING, "Unable to find peer subchannel for music on hold\n");
	break;
      }
    }
  return;
}

static void
KeyRinging (struct unistimsession *pte, char keycode)
{
  switch (keycode)
    {
    case KEY_HANGUP:
    case KEY_FUNC2:
      break;
    case KEY_FUNC4:
      IgnoreCall (pte);
      break;
    case KEY_FUNC1:
      HandleCallIncoming (pte);
      break;
    }
  return;
}

static void
KeyDialing (struct unistimsession *pte, char keycode)
{
  switch (keycode)
    {
    case KEY_HANGUP:
    case KEY_FUNC1:
      ShowMainPage (pte);
      CancelDial (pte);
      break;
    }
  return;
}

static void
KeyFavorite (struct unistimsession *pte, char keycode)
{
  int fav;

  if ((keycode < KEY_FAV1) && (keycode > KEY_FAV5))
    {
      ast_log (LOG_WARNING, "It's not a favorite key\n");
      return;
    }
  if (keycode == KEY_FAV0)
    return;
  fav = keycode - KEY_FAV0;
  if (pte->device->softkeyicon[fav] == 0)
    return;
  strncpy (pte->device->phone_number, pte->device->softkeynumber[fav], sizeof (pte->device->phone_number) - 1);
  HandleCallOutgoing (pte);
  return;
}
static void
SendDialTone (struct unistimsession *pte)
{
  int i;
  // BLUUU
  // No country defined ? Using US tone
  if (ast_strlen_zero (pte->device->country))
    {
      if (unistimdebug)
	ast_verbose ("No country defined, using US tone\n");
      SendTone (pte, 350, 440);
      return;
    }
  if (strlen (pte->device->country) != 2)
    {
      if (unistimdebug)
	ast_verbose ("Country code != 2 char, using US tone\n");
      SendTone (pte, 350, 440);
      return;
    }
  i = 0;
  while (frequency[i].freq1)
    {
      if ((frequency[i].country[0] == pte->device->country[0]) && (frequency[i].country[1] == pte->device->country[1]))
	{
	  if (unistimdebug)
	    ast_verbose ("Country code found (%s), freq1=%d freq2=%d\n", frequency[i].country, frequency[i].freq1,
			 frequency[i].freq2);
	  SendTone (pte, frequency[i].freq1, frequency[i].freq2);
	}
      i++;
    }
}

static void
HandleDialPage (struct unistimsession *pte)
{
  pte->state = STATE_DIALPAGE;
  if ((pte->device->output == OUTPUT_HANDSET) && (pte->device->receiver_state == STATE_ONHOOK))
    SendSelectOutput (pte, OUTPUT_SPEAKER, pte->device->volume, MUTE_OFF);
  else
    SendSelectOutput (pte, pte->device->output, pte->device->volume, MUTE_OFF);
  SendDialTone (pte);
  SendText (TEXT_LINE0, TEXT_NORMAL, pte, "Enter the number to dial");
  SendText (TEXT_LINE1, TEXT_NORMAL, pte, "and press Call");
  SendText (TEXT_LINE2, TEXT_INVERSE, pte, "Number : ...............");
  SendBlinkCursor (pte);
  SendCursorPos (pte, TEXT_LINE2 + 0x09);
  pte->device->size_phone_number = 0;
  pte->device->phone_number[0] = 0;
  SendTextStatus (pte, "Call   Redial BackSpcErase");
  ChangeFavoriteIcon (pte, FAV_ICON_PHONE_BLACK);
  SendIcon (TEXT_LINE0, FAV_ICON_NONE, pte);
  pte->device->missed_call = 0;
  SendLedUpdate (pte, 0);
  return;
}

static void
KeyDialPage (struct unistimsession *pte, char keycode)
{
  if (keycode == KEY_FUNC3)
    {
      if (pte->device->size_phone_number <= 1)
	keycode = KEY_FUNC4;
      else
	{
	  pte->device->size_phone_number -= 2;
	  keycode = pte->device->phone_number[pte->device->size_phone_number] + 0x10;
	}
    }
  if ((keycode >= KEY_0) && (keycode <= KEY_SHARP))
    {
      char tmpbuf[] = "Number : ...............";
      int i = 0;

      if (pte->device->size_phone_number >= 15)
	return;
      if (pte->device->size_phone_number == 0)
	SendTone (pte, 0, 0);
      while (i < pte->device->size_phone_number)
	{
	  tmpbuf[i + 9] = pte->device->phone_number[i];
	  i++;
	}
      if (keycode == KEY_SHARP)
	keycode = '#';
      else if (keycode == KEY_STAR)
	keycode = '*';
      else
	keycode -= 0x10;
      tmpbuf[i + 9] = keycode;
      pte->device->phone_number[i] = keycode;
      pte->device->size_phone_number++;
      pte->device->phone_number[i + 1] = 0;
      SendText (TEXT_LINE2, TEXT_INVERSE, pte, tmpbuf);
      SendBlinkCursor (pte);
      SendCursorPos (pte, (unsigned char) (TEXT_LINE2 + 0x0a + i));
      return;
    }

  switch (keycode)
    {
    case KEY_FUNC2:
      if (!strlen (pte->device->redial_number))
	break;
      strncpy (pte->device->phone_number, pte->device->redial_number, sizeof (pte->device->phone_number) - 1);
    case KEY_FUNC1:
      HandleCallOutgoing (pte);
      break;
    case KEY_FUNC4:
      pte->device->size_phone_number = 0;
      SendText (TEXT_LINE2, TEXT_INVERSE, pte, "Number : ...............");
      SendBlinkCursor (pte);
      SendCursorPos (pte, TEXT_LINE2 + 0x09);
      break;
    case KEY_HANGUP:
      ShowMainPage (pte);
      break;
    case KEY_FAV1:
    case KEY_FAV2:
    case KEY_FAV3:
    case KEY_FAV4:
    case KEY_FAV5:
      KeyFavorite (pte, keycode);
      break;
    case KEY_LOUDSPK:
      if (pte->device->output == OUTPUT_SPEAKER)
	{
	  if (pte->device->receiver_state == STATE_OFFHOOK)
	    SendSelectOutput (pte, pte->device->previous_output, pte->device->volume, MUTE_OFF);
	  else
	    ShowMainPage (pte);
	}
      else
	SendSelectOutput (pte, OUTPUT_SPEAKER, pte->device->volume, MUTE_OFF);
      break;
    case KEY_HEADPHN:
      if (pte->device->output == OUTPUT_HEADPHONE)
	{
	  if (pte->device->receiver_state == STATE_OFFHOOK)
	    SendSelectOutput (pte, OUTPUT_HANDSET, pte->device->volume, MUTE_OFF);
	  else
	    ShowMainPage (pte);
	}
      else
	SendSelectOutput (pte, OUTPUT_HEADPHONE, pte->device->volume, MUTE_OFF);
      break;
    }
  return;
}

#define SELECTCODEC_START_ENTRY_POS 15
#define SELECTCODEC_MAX_LENGTH 2
static void
HandleSelectCodec (struct unistimsession *pte)
{
  char buf[30], buf2[5];

  pte->state = STATE_SELECTCODEC;
  strcpy (buf, "Using codec ");
  sprintf (buf2, "%d", pte->device->codec_number);
  strcat (buf, buf2);
  strcat (buf, " (G711u=0,");

  SendText (TEXT_LINE0, TEXT_NORMAL, pte, buf);
  SendText (TEXT_LINE1, TEXT_NORMAL, pte, "G723=4,G711a=8,G729A=18)");
  SendText (TEXT_LINE2, TEXT_INVERSE, pte, "Codec number : ..");
  SendBlinkCursor (pte);
  SendCursorPos (pte, TEXT_LINE2 + SELECTCODEC_START_ENTRY_POS);
  pte->device->size_buff_entry = 0;
  SendTextStatus (pte, "Select BackSpcErase  Cancel");
  return;
}

static void
KeySelectCodec (struct unistimsession *pte, char keycode)
{
  if (keycode == KEY_FUNC2)
    {
      if (pte->device->size_buff_entry <= 1)
	keycode = KEY_FUNC3;
      else
	{
	  pte->device->size_buff_entry -= 2;
	  keycode = pte->device->buff_entry[pte->device->size_buff_entry] + 0x10;
	}
    }
  if ((keycode >= KEY_0) && (keycode <= KEY_9))
    {
      char tmpbuf[] = "Codec number : ..";
      int i = 0;

      if (pte->device->size_buff_entry >= SELECTCODEC_MAX_LENGTH)
	return;

      while (i < pte->device->size_buff_entry)
	{
	  tmpbuf[i + SELECTCODEC_START_ENTRY_POS] = pte->device->buff_entry[i];
	  i++;
	}
      tmpbuf[i + SELECTCODEC_START_ENTRY_POS] = keycode - 0x10;
      pte->device->buff_entry[i] = keycode - 0x10;
      pte->device->size_buff_entry++;
      SendText (TEXT_LINE2, TEXT_INVERSE, pte, tmpbuf);
      SendBlinkCursor (pte);
      SendCursorPos (pte, (unsigned char) (TEXT_LINE2 + SELECTCODEC_START_ENTRY_POS + 1 + i));
      return;
    }

  switch (keycode)
    {
    case KEY_FUNC1:
      if (pte->device->size_buff_entry == 1)
	pte->device->codec_number = pte->device->buff_entry[0] - 48;
      else if (pte->device->size_buff_entry == 2)
	pte->device->codec_number = ((pte->device->buff_entry[0] - 48) * 10) + (pte->device->buff_entry[1] - 48);
      ShowMainPage (pte);
      break;
    case KEY_FUNC3:
      pte->device->size_buff_entry = 0;
      SendText (TEXT_LINE2, TEXT_INVERSE, pte, "Codec number : ..");
      SendBlinkCursor (pte);
      SendCursorPos (pte, TEXT_LINE2 + SELECTCODEC_START_ENTRY_POS);
      break;
    case KEY_HANGUP:
    case KEY_FUNC4:
      ShowMainPage (pte);
      break;
    }
  return;
}

static void
ShowMainPage (struct unistimsession *pte)
{
  char tmpbuf[TEXT_LENGTH_MAX + 1];
  char iabuf[INET_ADDRSTRLEN];
  pte->state = STATE_MAINPAGE;

  SendTone (pte, 0, 0);
  SendSelectOutput (pte, pte->device->output, pte->device->volume, MUTE_ON_DISCRET);
  pte->device->lines->lastmsgssent = 0;
  SendFavorite (0, FAV_ICON_ONHOOK_BLACK, pte, pte->device->softkeylabel[0]);
  if (pte->device->missed_call == 0)
    SendText (TEXT_LINE0, TEXT_NORMAL, pte, pte->device->maintext0);
  else
    {
      sprintf (tmpbuf, "%d unanswered call(s)", pte->device->missed_call);
      SendText (TEXT_LINE0, TEXT_NORMAL, pte, tmpbuf);
      SendIcon (TEXT_LINE0, FAV_ICON_CALL_CENTER + FAV_BLINK_SLOW, pte);
    }
  SendText (TEXT_LINE1, TEXT_NORMAL, pte, pte->device->maintext1);
  if (!strlen (pte->device->maintext2))
    {
      strcpy (tmpbuf, "IP : ");
      strcat (tmpbuf, ast_inet_ntoa (iabuf, sizeof (iabuf), pte->sin.sin_addr));
      SendText (TEXT_LINE2, TEXT_NORMAL, pte, tmpbuf);
    }
  else
    SendText (TEXT_LINE2, TEXT_NORMAL, pte, pte->device->maintext2);
  SendTextStatus (pte, "Dial   Redial");
  SendTextTitle (pte, pte->device->titledefault);
  ChangeFavoriteIcon (pte, FAV_ICON_ONHOOK_BLACK);
}

static void
KeyMainPage (struct unistimsession *pte, char keycode)
{
  if (pte->device->missed_call)
    {
      SendIcon (TEXT_LINE0, FAV_ICON_NONE, pte);
      pte->device->missed_call = 0;
    }
  if ((keycode >= KEY_0) && (keycode <= KEY_SHARP))
    {
      HandleDialPage (pte);
      KeyDialPage (pte, keycode);
      return;
    }
  switch (keycode)
    {
    case KEY_FUNC1:
      HandleDialPage (pte);
      break;
    case KEY_FUNC2:
      if (!strlen (pte->device->redial_number))
	break;
      if ((pte->device->output == OUTPUT_HANDSET) && (pte->device->receiver_state == STATE_ONHOOK))
	SendSelectOutput (pte, OUTPUT_SPEAKER, pte->device->volume, MUTE_OFF);
      else
	SendSelectOutput (pte, pte->device->output, pte->device->volume, MUTE_OFF);

      strncpy (pte->device->phone_number, pte->device->redial_number, sizeof (pte->device->phone_number) - 1);
      HandleCallOutgoing (pte);
      break;
    case KEY_FAV0:
      HandleDialPage (pte);
      break;
    case KEY_FAV1:
    case KEY_FAV2:
    case KEY_FAV3:
    case KEY_FAV4:
    case KEY_FAV5:
      if ((pte->device->output == OUTPUT_HANDSET) && (pte->device->receiver_state == STATE_ONHOOK))
	SendSelectOutput (pte, OUTPUT_SPEAKER, pte->device->volume, MUTE_OFF);
      else
	SendSelectOutput (pte, pte->device->output, pte->device->volume, MUTE_OFF);
      KeyFavorite (pte, keycode);
      break;
    case KEY_CONF:
      HandleSelectCodec (pte);
      break;
    case KEY_LOUDSPK:
      SendSelectOutput (pte, OUTPUT_SPEAKER, pte->device->volume, MUTE_OFF);
      HandleDialPage (pte);
      break;
    case KEY_HEADPHN:
      SendSelectOutput (pte, OUTPUT_HEADPHONE, pte->device->volume, MUTE_OFF);
      HandleDialPage (pte);
      break;
    }
  return;
}

static void
InitPhoneStep2 (struct unistimsession *pte)
{
  if (unistimdebug)
    ast_verbose ("Sending S4\n");
  memcpy (buffsend + SIZE_HEADER, PacketSendS4, sizeof (PacketSendS4));
  SendClient (SIZE_HEADER + sizeof (PacketSendS4), buffsend, pte);
  SendDateTime2 (pte);
  SendDateTime3 (pte);
  if (unistimdebug)
    ast_verbose ("Sending S7\n");
  memcpy (buffsend + SIZE_HEADER, PacketSendS7, sizeof (PacketSendS7));
  SendClient (SIZE_HEADER + sizeof (PacketSendS7), buffsend, pte);
  if (unistimdebug)
    ast_verbose ("Sending Contrast\n");
  memcpy (buffsend + SIZE_HEADER, PacketSendContrast, sizeof (PacketSendContrast));
  if (pte->device != NULL)
    buffsend[9] = pte->device->contrast;
  SendClient (SIZE_HEADER + sizeof (PacketSendContrast), buffsend, pte);

  if (unistimdebug)
    ast_verbose ("Sending S9\n");
  memcpy (buffsend + SIZE_HEADER, PacketSendS9, sizeof (PacketSendS9));
  SendClient (SIZE_HEADER + sizeof (PacketSendS9), buffsend, pte);
  SendNoRing (pte);

  if (unistimdebug)
    ast_verbose ("Sending S7\n");
  memcpy (buffsend + SIZE_HEADER, PacketSendS7, sizeof (PacketSendS7));
  SendClient (SIZE_HEADER + sizeof (PacketSendS7), buffsend, pte);
  SendLedUpdate (pte, 0);
  SendPing (pte);
  if (pte->state < STATE_MAINPAGE)
    {
      int i;
      char tmp[30];

      for (i = 1; i < 6; i++)
	SendFavorite (i, 0, pte, "");
      SendText (TEXT_LINE0, TEXT_NORMAL, pte, "Sorry, this phone is not");
      SendText (TEXT_LINE1, TEXT_NORMAL, pte, "registred in unistim.cfg");
      strcpy (tmp, "MAC = ");
      strcat (tmp, pte->macaddr);
      SendText (TEXT_LINE2, TEXT_NORMAL, pte, tmp);
      SendTextStatus (pte, "");
      SendTextTitle (pte, "UNISTIM for*");
      return;
    }
  else
    ShowMainPage (pte);
  if (unistimdebug)
    ast_verbose ("Sending Arrow\n");
  memcpy (buffsend + SIZE_HEADER, PacketSendArrow, sizeof (PacketSendArrow));
  SendClient (SIZE_HEADER + sizeof (PacketSendArrow), buffsend, pte);
  RefreshAllFavorite (pte);
  return;
}

static void
process_request (int size, unsigned char *buf, struct unistimsession *pte)
{
  char tmpbuf[255];
  char iabuf[INET_ADDRSTRLEN];
  if (memcmp (buf + SIZE_HEADER, PacketRecvResumeConnectionWithServer, sizeof (PacketRecvResumeConnectionWithServer)) == 0)
    {
      RcvResumeConnectionWithServer (pte);
      return;
    }
  if (memcmp (buf + SIZE_HEADER, PacketRecvFirmVersion, sizeof (PacketRecvFirmVersion)) == 0)
    {
      buf[size] = 0;
      if (unistimdebug)
	ast_verbose ("Got the firmware version : '%s'\n", buf + 13);
      InitPhoneStep2 (pte);
      return;
    }
  if (memcmp (buf + SIZE_HEADER, PacketRecvMacAddr, sizeof (PacketRecvMacAddr)) == 0)
    {
      RcvMacAddr (pte, buf);
      return;
    }
  if (memcmp (buf + SIZE_HEADER, PacketRecvR2, sizeof (PacketRecvR2)) == 0)
    {
      if (unistimdebug)
	ast_verbose ("R2 received\n");
      return;
    }

  if (pte->state < STATE_MAINPAGE)
    {
      if (unistimdebug)
	ast_verbose ("Request not authorized in this state\n");
      return;
    }
  if (memcmp (buf + SIZE_HEADER, PacketRecvPressedKey, sizeof (PacketRecvPressedKey)) == 0)
    {
      char keycode = buf[13];

      if (unistimdebug)
	ast_verbose ("Key pressed : keycode = 0x%.2x - current state : %d\n", keycode, pte->state);
      // test key
      if (keycode == KEY_COMPUTR)
	{
	  SendTest (pte);
	  return;
	}
      // End of the test section
      switch (pte->state)
	{
	case STATE_INIT:
	  if (unistimdebug)
	    ast_verbose ("No keys allowed in the init state\n");
	  break;
	case STATE_AUTHDENY:
	  if (unistimdebug)
	    ast_verbose ("No keys allowed in authdeny state\n");
	  break;
	case STATE_MAINPAGE:
	  KeyMainPage (pte, keycode);
	  break;
	case STATE_DIALPAGE:
	  KeyDialPage (pte, keycode);
	  break;
	case STATE_DIALING:
	  KeyDialing (pte, keycode);
	  break;
	case STATE_RINGING:
	  KeyRinging (pte, keycode);
	  break;
	case STATE_CALL:
	  KeyCall (pte, keycode);
	  break;
	case STATE_SELECTCODEC:
	  KeySelectCodec (pte, keycode);
	  break;
	default:
	  ast_log (LOG_WARNING, "Key : Unknown state\n");
	}
      return;
    }
  if (memcmp (buf + SIZE_HEADER, PacketRecvPickUp, sizeof (PacketRecvPickUp)) == 0)
    {
      if (unistimdebug)
	ast_verbose ("Handset off hook\n");
      pte->device->receiver_state = STATE_OFFHOOK;
      if (pte->device->output == OUTPUT_HEADPHONE)
	SendSelectOutput (pte, OUTPUT_HEADPHONE, pte->device->volume, MUTE_OFF);
      else
	SendSelectOutput (pte, OUTPUT_HANDSET, pte->device->volume, MUTE_OFF);
      if (pte->state == STATE_RINGING)
	HandleCallIncoming (pte);
      else if ((pte->state == STATE_DIALPAGE) || (pte->state == STATE_CALL))
	SendSelectOutput (pte, OUTPUT_HANDSET, pte->device->volume, MUTE_OFF);
      else if (pte->state != STATE_DIALING)
	{
	  SendSelectOutput (pte, OUTPUT_HANDSET, pte->device->volume, MUTE_OFF);
	  HandleDialPage (pte);
	}
      return;
    }
  if (memcmp (buf + SIZE_HEADER, PacketRecvHangUp, sizeof (PacketRecvHangUp)) == 0)
    {
      if (unistimdebug)
	ast_verbose ("Handset on hook\n");
      pte->device->receiver_state = STATE_ONHOOK;
      if (pte->state == STATE_DIALING)
	CancelDial (pte);
      else if (pte->state == STATE_CALL)
	CloseCall (pte);
      else
	ShowMainPage (pte);
      return;
    }
  strcpy (tmpbuf, ast_inet_ntoa (iabuf, sizeof (iabuf), pte->sin.sin_addr));
  strcat (tmpbuf, " Unknown request packet\n");
  if (unistimdebug)
    ast_log (LOG_DEBUG, tmpbuf);
  return;
}

static void
parsing (int size, unsigned char *buf, struct unistimsession *pte, struct sockaddr_in *addr_from)
{
  unsigned short *sbuf = (unsigned short *) buf;
  unsigned short seq;
  char tmpbuf[255];
  char iabuf[INET_ADDRSTRLEN];

  strcpy (tmpbuf, ast_inet_ntoa (iabuf, sizeof (iabuf), addr_from->sin_addr));

  if (size < 10)
    {
      if (size == 0)
	{
	  ast_log (LOG_WARNING, "%s Read error\n", tmpbuf);
	}
      else
	{
	  ast_log (LOG_NOTICE, "%s Packet too short - ignoring\n", tmpbuf);
	}
      return;
    }
  if (sbuf[0] == 0xffff)	// Starting with 0xffff ?
    {				// Yes, discovery packet ?
      if (size != sizeof (PacketRcvDiscovery))
	{
	  ast_log (LOG_NOTICE, "%s Invalid size of a discovery packet\n", tmpbuf);
	}
      else
	{
	  if (memcmp (buf, PacketRcvDiscovery, sizeof (PacketRcvDiscovery)) == 0)
	    {
	      if (unistimdebug)
		ast_verbose ("Discovery packet received - Sending Discovery ACK\n");
	      if (pte)		/* A session was already active for this IP ? */
		{
		  if (pte->state == STATE_INIT)	/* Yes, but it's a dupe */
		    {
		      if (unistimdebug)
			ast_verbose ("Duplicated Discovery packet\n");
		      SendRawClient (sizeof (PacketSendDiscoveryACK), PacketSendDiscoveryACK, addr_from, &pte->sout);
		      pte->seq_phone = (short) 0x0000;	// reset sequence number
		    }
		  else
		    {		/* No, probably a reboot, phone side */
		      CloseClient (pte);	/* Cleanup the previous session */
		      if (CreateClient (addr_from))
			SendRawClient (sizeof (PacketSendDiscoveryACK), PacketSendDiscoveryACK, addr_from, &pte->sout);
		    }
		}
	      else
		{
		  // Creating new entry in our phone list
		  if ((pte = CreateClient (addr_from)))
		    SendRawClient (sizeof (PacketSendDiscoveryACK), PacketSendDiscoveryACK, addr_from, &pte->sout);
		}
	      return;
	    }
	  ast_log (LOG_NOTICE, "%s Invalid discovery packet\n", tmpbuf);
	}
      return;
    }
  if (!pte)
    {
      if (unistimdebug)
	ast_verbose ("%s Not a discovery packet from an unknown source : ignoring\n", tmpbuf);
      return;
    }

  if (sbuf[0] != 0)		// Starting with something else than 0x0000 ?
    {
      ast_log (LOG_NOTICE, "Unknown packet received - ignoring\n");
      return;
    }
  if (buf[5] != 2)
    {
      ast_log (LOG_NOTICE, "%s Wrong direction : got 0x%.2x expected 0x02\n", tmpbuf, buf[5]);
      return;
    }
  seq = ntohs (sbuf[1]);
  if (buf[4] == 1)
    {
      if ((unistimdebug) && (option_verbose > 5))
	ast_verbose ("ACK received for packet #0x%.4x\n", seq);
      pte->nb_retransmit = 0;

      if ((pte->last_seq_ack) + 1 == seq)
	{
	  pte->last_seq_ack++;
	  if (pte->last_buf_available == 1)
	    {
	      // Buffer empty - setting ping timer
	      pte->last_buf_available--;
	      SetPingTimer (pte);
	    }
	  else if (pte->last_seq_ack + 1 == pte->seq_server + 1)
	    {
	      // Buffer empty - setting ping timer
	      pte->last_buf_available = 0;
	      SetPingTimer (pte);
	    }
	  return;
	}
      if (pte->last_seq_ack > seq)
	{
	  ast_log (LOG_NOTICE, "%s Warning : ACK received for an already ACKed packet : #0x%.4x\n", tmpbuf, pte->last_seq_ack);
	  return;
	}
      if (pte->seq_server < seq)
	{
	  ast_log (LOG_NOTICE, "%s Error : ACK received a non-existant packet : #0x%.4x\n", tmpbuf, pte->seq_server);
	  return;
	}
      ast_log (LOG_NOTICE, "%s Warning : ACK lost\n", tmpbuf);
      pte->last_seq_ack = seq;
      return;
    }
  if (buf[4] == 2)
    {
      if (unistimdebug)
	ast_verbose ("Request received\n");
      if (pte->seq_phone == seq)
	{
	  // Send ACK
	  buf[4] = 1;
	  buf[5] = 1;
	  SendRawClient (SIZE_HEADER, buf, addr_from, &pte->sout);
	  pte->seq_phone++;

	  process_request (size, buf, pte);
	  return;
	}
      if (pte->seq_phone > seq)
	{
	  ast_log (LOG_NOTICE,
		   "%s Warning : received a retransmitted packet : #0x%.4x (we are at #0x%.4x)\n", tmpbuf, seq, pte->seq_phone);
	  // BUG ? pte->device->seq_phone = seq;
	  // Send ACK
	  buf[4] = 1;
	  buf[5] = 1;
	  SendRawClient (SIZE_HEADER, buf, addr_from, &pte->sout);
	  return;
	}
      ast_log (LOG_NOTICE, "%s Warning : we lost a packet : received #0x%.4x (we are at #0x%.4x)\n", tmpbuf, seq, pte->seq_phone);
      return;
    }
  if (buf[4] == 0)
    {
      ast_log (LOG_NOTICE, "%s Retransmit request for packet #0x%.4x\n", tmpbuf, seq);
      if (pte->last_seq_ack > seq)
	{
	  ast_log (LOG_NOTICE, "%s Error : received a request for an already ACKed packet : #0x%.4x\n", tmpbuf, pte->last_seq_ack);
	  return;
	}
      if (pte->seq_server < seq)
	{
	  ast_log (LOG_NOTICE, "%s Error : received a request for a non-existant packet : #0x%.4x\n", tmpbuf, pte->seq_server);
	  return;
	}
      SendRetransmit (pte);
      return;
    }
  ast_log (LOG_NOTICE, "%s Unknown request : got 0x%.2x expected 0x00,0x01 or 0x02\n", tmpbuf, buf[4]);
  return;
}

/*--- unistim_call: Initiate UNISTIM call from PBX ---*/
/*      used from the dial() application      */
static int
unistim_call (struct ast_channel *ast, char *dest, int timeout)
{
  int res = 0;
  struct unistim_line *l;
  struct unistim_subchannel *sub;
  struct unistimsession *session;
  char cid[256];
  char *lc = NULL, *n = NULL;

  sub = ast->tech_pvt;
  l = sub->parent;
  session = l->parent->session;

  if (!l->parent->session)
    {
      ast_log (LOG_ERROR, "Device not registered, cannot call %s\n", dest);
      return -1;
    }

  if ((ast->_state != AST_STATE_DOWN) && (ast->_state != AST_STATE_RESERVED))
    {
      ast_log (LOG_WARNING, "unistim_call called on %s, neither down nor reserved\n", ast->name);
      return -1;
    }

  if (unistimdebug)
    ast_verbose (VERBOSE_PREFIX_3 "unistim_call(%s)\n", ast->name);

  if (l->dnd)
    {
      ast_queue_control (ast, AST_CONTROL_BUSY);
      return -1;
    }

  session->state = STATE_RINGING;
  SendIcon (TEXT_LINE0, FAV_ICON_NONE, session);

  if (sub->owner && sub->owner->cid.cid_num)
    {
      strncpy (cid, sub->owner->cid.cid_num, sizeof (cid) - 1);
      cid[sizeof (cid) - 1] = '\0';
      ast_callerid_parse (cid, &n, &lc);
    }
  if (!n)
    n = default_callername;
  if (!lc)
    lc = default_callerid;
  SendText (TEXT_LINE0, TEXT_NORMAL, session, n);
  SendText (TEXT_LINE1, TEXT_NORMAL, session, lc);
  SendText (TEXT_LINE2, TEXT_NORMAL, session, "is calling you.");
  SendTextStatus (session, "Accept               Ignore");

  if (sub->ringstyle == -1)
    SendRing (session, session->device->ringvolume, session->device->ringstyle);
  else
    {
      if (sub->ringvolume == -1)
	SendRing (session, session->device->ringvolume, sub->ringstyle);
      else
	SendRing (session, sub->ringvolume, sub->ringstyle);
    }
  ChangeFavoriteIcon (session, FAV_ICON_SPEAKER_ONHOOK_BLACK + FAV_BLINK_FAST);

  ast_setstate (ast, AST_STATE_RINGING);
  ast_queue_control (ast, AST_CONTROL_RINGING);
  return res;
}

/*--- unistim_hangup: Hangup UNISTIM call */
static int
unistim_hangup (struct ast_channel *ast)
{
  struct unistim_subchannel *sub = ast->tech_pvt;
  struct unistim_line *l = sub->parent;
  struct unistimsession *s = l->parent->session;

  if (unistimdebug)
    ast_verbose ("unistim_hangup(%s) on %s@%s\n", ast->name, l->name, l->parent->name);
  if (!ast->tech_pvt)
    {
      ast_log (LOG_DEBUG, "Asked to hangup channel not connected (no subchannel)\n");
      return 0;
    }
  ast_mutex_lock (&sub->lock);
  sub->owner = NULL;
  ast->tech_pvt = NULL;
  sub->alreadygone = 0;
  sub->outgoing = 0;
  sub->ringing = 0;
  ast_mutex_unlock (&sub->lock);
  if (!s)
    {
      if (unistimdebug)
	ast_verbose ("Asked to hangup channel not connected (no session)\n");
      if (sub->rtp)
	{
	  if (unistimdebug)
	    ast_verbose ("Destroying RTP session\n");
	  ast_rtp_destroy (sub->rtp);
	  sub->rtp = NULL;
	}

      return 0;
    }
  SendNoRing (s);
  if ((s->state == STATE_DIALING) || (s->state == STATE_RINGING))
    CancelDial (s);
  else if (s->state == STATE_CALL)
    CloseCall (s);
  if (sub->rtp)
    {
      ast_rtp_destroy (sub->rtp);
      sub->rtp = NULL;
    }
  return 0;
}

/*--- unistim_answer: Answer UNISTIM call */
static int
unistim_answer (struct ast_channel *ast)
{
  int res = 0;
  struct unistim_subchannel *sub = ast->tech_pvt;
  struct unistim_line *l = sub->parent;
  if (!sub->rtp)
    start_rtp (sub);
  if (unistimdebug)
    ast_verbose ("unistim_answer(%s) on %s@%s-%d\n", ast->name, l->name, l->parent->name, sub->callid);
  if (!l->parent->session)
    ast_log (LOG_WARNING, "unistim_answer on a disconnected device ?\n");
  else
    {
      SendText (TEXT_LINE2, TEXT_NORMAL, l->parent->session, "is now on-line");
      SendStartTimer (l->parent->session);
    }
  if (ast->_state != AST_STATE_UP)
    ast_setstate (ast, AST_STATE_UP);

  return res;
}

/*--- unistimsock_read: Read data from UNISTIM socket ---*/
/*    Successful messages is connected to UNISTIM call and forwarded to parsing() */
static int
unistimsock_read (int *id, int fd, short events, void *ignore)
{
  struct sockaddr_in addr_from = { 0, };
  struct unistimsession *cur = NULL;
  int found = 0;
  int tmp = 0;
  unsigned int dwNumberOfBytesRecv;
#ifdef DUMP_PACKET
  unsigned int dwNumberOfBytesRecvd;
  char iabuf[INET_ADDRSTRLEN];
#endif

  dwNumberOfBytesRecv = recvfrom (unistimsock, buff, SIZE_PAGE, 0, (struct sockaddr *) &addr_from, &size_addr_from);
  if (dwNumberOfBytesRecv == -1)
    {
      if (errno == EAGAIN)
	ast_log (LOG_NOTICE, "UNISTIM: Received packet with bad UDP checksum\n");
      else if (errno != ECONNREFUSED)
	ast_log (LOG_WARNING, "Recv error %d (%s)\n", errno, strerror (errno));
      return 1;
    }

  // Looking in the phone list if we already have a registration for him
  ast_mutex_lock (&sessionlock);
  cur = sessions;
  while (cur)
    {
      if (cur->sin.sin_addr.s_addr == addr_from.sin_addr.s_addr)
	{
	  found = 1;
	  break;
	}
      tmp++;
      cur = cur->next;
    }
  ast_mutex_unlock (&sessionlock);

#ifdef DUMP_PACKET
  if (unistimdebug)
    ast_verbose ("\n*** Dump %d bytes from %s - phone_table[%d] ***\n",
		 dwNumberOfBytesRecv, ast_inet_ntoa (iabuf, sizeof (iabuf), addr_from.sin_addr), tmp);
  for (dwNumberOfBytesRecvd = 0; dwNumberOfBytesRecvd < dwNumberOfBytesRecv; dwNumberOfBytesRecvd++)
    ast_verbose ("%.2x ", (unsigned char) buff[dwNumberOfBytesRecvd]);
  ast_verbose ("\n******************************************\n");
#endif

  if (!found)
    {
      if (unistimdebug)
	ast_verbose ("Received a packet from an unknown source\n");
      parsing (dwNumberOfBytesRecv, buff, NULL, (struct sockaddr_in *) &addr_from);

    }
  else
    parsing (dwNumberOfBytesRecv, buff, cur, (struct sockaddr_in *) &addr_from);

  return 1;
}
static struct ast_frame *
unistim_rtp_read (struct ast_channel *ast, struct unistim_subchannel *sub)
{
  /* Retrieve audio/etc from channel.  Assumes sub->lock is already held. */
  struct ast_frame *f;
  static struct ast_frame null_frame = { AST_FRAME_NULL, };
  if ((!ast) || (!sub->rtp))
    {
      ast_log (LOG_WARNING, "Channel or RTP handle NULL while reading\n");
      return &null_frame;
    }
  switch (ast->fdno)
    {
    case 0:
      f = ast_rtp_read (sub->rtp);	/* RTP Audio */
      break;
    case 1:
      f = ast_rtcp_read (sub->rtp);	/* RTCP Control Channel */
      break;
    default:
      f = &null_frame;
    }
  if (sub->owner)
    {
      /* We already hold the channel lock */
      if (f->frametype == AST_FRAME_VOICE)
	{
	  if (f->subclass != sub->owner->nativeformats)
	    {
	      ast_log (LOG_DEBUG, "Oooh, format changed from %s (%d) to %s (%d)\n",
		       ast_getformatname (sub->owner->nativeformats), sub->owner->nativeformats,
		       ast_getformatname (f->subclass), f->subclass);

	      sub->owner->nativeformats = f->subclass;
	      ast_set_read_format (sub->owner, sub->owner->readformat);
	      ast_set_write_format (sub->owner, sub->owner->writeformat);
	    }
	}
    }
  return f;
}

static struct ast_frame *
unistim_read (struct ast_channel *ast)
{
  struct ast_frame *fr;
  struct unistim_subchannel *sub = ast->tech_pvt;
  ast_mutex_lock (&sub->lock);
  fr = unistim_rtp_read (ast, sub);
  ast_mutex_unlock (&sub->lock);
  return fr;
}

static int
unistim_write (struct ast_channel *ast, struct ast_frame *frame)
{
  struct unistim_subchannel *sub = ast->tech_pvt;
  int res = 0;
  if (frame->frametype != AST_FRAME_VOICE)
    {
      if (frame->frametype == AST_FRAME_IMAGE)
	return 0;
      else
	{
	  ast_log (LOG_WARNING, "Can't send %d type frames with unistim_write\n", frame->frametype);
	  return 0;
	}
    }
  else
    {
      if (!(frame->subclass & ast->nativeformats))
	{
	  ast_log (LOG_WARNING,
		   "Asked to transmit frame type %s (%d), while native formats is %s (%d) (read/write = %s (%d)/%d)\n",
		   ast_getformatname (frame->subclass), frame->subclass, ast_getformatname (ast->nativeformats),
		   ast->nativeformats, ast_getformatname (ast->readformat), ast->readformat, ast->writeformat);
	  return -1;
	}
    }
  if (sub)
    {
      ast_mutex_lock (&sub->lock);
      if (sub->rtp)
	{
	  res = ast_rtp_write (sub->rtp, frame);
	}
      ast_mutex_unlock (&sub->lock);
    }
  return res;
}

static int
unistim_fixup (struct ast_channel *oldchan, struct ast_channel *newchan)
{
  struct unistim_subchannel *sub = newchan->tech_pvt;
  ast_log (LOG_NOTICE, "unistim_fixup(%s, %s)\n", oldchan->name, newchan->name);
  if (sub->owner != oldchan)
    {
      ast_log (LOG_WARNING, "old channel wasn't %p but was %p\n", oldchan, sub->owner);
      return -1;
    }
  sub->owner = newchan;
  return 0;
}

static char *
control2str (int ind)
{
  switch (ind)
    {
    case AST_CONTROL_HANGUP:
      return "Other end has hungup";
    case AST_CONTROL_RING:
      return "Local ring";
    case AST_CONTROL_RINGING:
      return "Remote end is ringing";
    case AST_CONTROL_ANSWER:
      return "Remote end has answered";
    case AST_CONTROL_BUSY:
      return "Remote end is busy";
    case AST_CONTROL_TAKEOFFHOOK:
      return "Make it go off hook";
    case AST_CONTROL_OFFHOOK:
      return "Line is off hook";
    case AST_CONTROL_CONGESTION:
      return "Congestion (circuits busy)";
    case AST_CONTROL_FLASH:
      return "Flash hook";
    case AST_CONTROL_WINK:
      return "Wink";
    case AST_CONTROL_OPTION:
      return "Set a low-level option";
    case AST_CONTROL_RADIO_KEY:
      return "Key Radio";
    case AST_CONTROL_RADIO_UNKEY:
      return "Un-Key Radio";
    case -1:
      return "Stop tone";
    }
  return "UNKNOWN";
}

static int
unistim_indicate (struct ast_channel *ast, int ind)
{
  struct unistim_subchannel *sub = ast->tech_pvt;
  struct unistim_line *l = sub->parent;
  struct unistimsession *s = l->parent->session;

  if (unistimdebug)
    ast_verbose (VERBOSE_PREFIX_3 "Asked to indicate '%s' condition on channel %s\n", control2str (ind), ast->name);
  switch (ind)
    {
    case AST_CONTROL_RINGING:
      if (ast->_state != AST_STATE_UP)
	{
	  if (!sub->progress)
	    {
	      sub->ringing = 1;
	      SendText (TEXT_LINE2, TEXT_NORMAL, s, "Ringing...");
	      break;
	    }
	}
      return -1;
    case AST_CONTROL_BUSY:
      if (ast->_state != AST_STATE_UP)
	{
	  SendTone (s, 480, 620);
	  sub->alreadygone = 1;
	  ast_softhangup_nolock (ast, AST_SOFTHANGUP_DEV);
	  break;
	}
      return -1;
    case AST_CONTROL_CONGESTION:
      if (ast->_state != AST_STATE_UP)
	{
	  SendTone (s, 440, 250);
	  sub->alreadygone = 1;
	  ast_softhangup_nolock (ast, AST_SOFTHANGUP_DEV);
	  break;
	}
      return -1;
    case AST_CONTROL_PROGRESS:
      if ((ast->_state != AST_STATE_UP) && !sub->progress && !sub->outgoing)
	{
	  sub->progress = 1;
	  break;
	}
      return -1;
    case -1:
      break;
    case AST_CONTROL_PROCEEDING:
      break;
    default:
      ast_log (LOG_WARNING, "Don't know how to indicate condition %d\n", ind);
      return -1;
    }
  return 0;
}

static struct unistim_subchannel *
find_subchannel_by_name (char *dest)
{
  struct unistim_line *l;
  struct unistim_device *d;
  char line[256];
  char *at;
  char *device;

  strncpy (line, dest, sizeof (line) - 1);
  at = strchr (line, '@');
  if (!at)
    {
      ast_log (LOG_NOTICE, "Device '%s' has no @ (at) sign!\n", dest);
      return NULL;
    }
  *at = '\0';
  at++;
  device = at;
  ast_mutex_lock (&devicelock);
  d = devices;
  at = strchr (device, '/');	// Extra options ?
  if (at)
    *at = '\0';
  while (d)
    {
      if (!strcasecmp (d->name, device))
	{
	  if (unistimdebug)
	    ast_verbose ("Found device: %s\n", d->name);
	  /* Found the device */
	  l = d->lines;
	  while (l)
	    {
	      /* Search for the right line */
	      if (!strcasecmp (l->name, line))
		{
		  l->sub->ringvolume = -1;
		  l->sub->ringstyle = -1;
		  if (at)	// Other options ?
		    {
		      at++;	// Skip slash
		      if (*at == 'r')	// distinctive ring
			{
			  at++;
			  if ((*at < '0') || (*at > '7'))	// ring style
			    ast_log (LOG_WARNING, "Invalid ring selection (%s)", at);
			  else
			    {
			      char ring_volume = -1;
			      char ring_style = *at - '0';
			      at++;
			      if ((*at >= '0') && (*at <= '3'))	// ring volume
				ring_volume = *at - '0';
			      if (unistimdebug)
				ast_verbose ("Distinctive ring : style #%d volume %d\n", ring_style, ring_volume);
			      l->sub->ringvolume = ring_volume;
			      l->sub->ringstyle = ring_style;
			    }
			}
		    }
		  ast_mutex_unlock (&devicelock);
		  return l->sub;
		}
	      l = l->next;
	    }
	}
      d = d->next;
    }
  /* Device not found */
  ast_mutex_unlock (&devicelock);
  return NULL;
}

/* We can do something with the tone generator */
static int
unistim_senddigit (struct ast_channel *ast, char digit)
{
  return -1;
}

/*--- unistim_sendtext: Display a text on the phone screen ---*/
/*      Called from PBX core text message functions */
static int
unistim_sendtext (struct ast_channel *ast, const char *text)
{
  struct unistim_subchannel *sub;
  struct unistimsession *pte;
  int size;
  char tmp[TEXT_LENGTH_MAX + 1];
  if (unistimdebug)
    ast_verbose ("unistim_sendtext called\n");
  sub = ast->tech_pvt;
  pte = sub->parent->parent->session;
  size = strlen (text);
  if (size <= TEXT_LENGTH_MAX * 2)
    {
      SendText (TEXT_LINE0, TEXT_NORMAL, pte, "Message :");
      SendText (TEXT_LINE1, TEXT_NORMAL, pte, text);
      if (size <= TEXT_LENGTH_MAX)
	{
	  SendText (TEXT_LINE2, TEXT_NORMAL, pte, "");
	  return 0;
	}
      memcpy (tmp, text + TEXT_LENGTH_MAX, TEXT_LENGTH_MAX);
      tmp[TEXT_LENGTH_MAX + 1] = '\0';
      SendText (TEXT_LINE2, TEXT_NORMAL, pte, tmp);
      return 0;
    }
  SendText (TEXT_LINE0, TEXT_NORMAL, pte, text);
  memcpy (tmp, text + TEXT_LENGTH_MAX, TEXT_LENGTH_MAX);
  tmp[TEXT_LENGTH_MAX + 1] = '\0';
  SendText (TEXT_LINE1, TEXT_NORMAL, pte, tmp);
  memcpy (tmp, text + TEXT_LENGTH_MAX * 2, TEXT_LENGTH_MAX);
  tmp[TEXT_LENGTH_MAX + 1] = '\0';
  SendText (TEXT_LINE2, TEXT_NORMAL, pte, tmp);
  return 0;
}

/*--- unistim_send_mwi_to_peer: Send message waiting indication ---*/
static int
unistim_send_mwi_to_peer (struct unistimsession *s)
{
  int newmsgs, oldmsgs;
  struct unistim_line *peer = s->device->lines;

  /* Check for messages */
  ast_app_messagecount (peer->mailbox, &newmsgs, &oldmsgs);
  time (&peer->lastmsgcheck);

  /* Return now if it's the same thing we told them last time */
  if (((newmsgs << 8) | (oldmsgs)) == peer->lastmsgssent)
    return 0;

  peer->lastmsgssent = ((newmsgs << 8) | (oldmsgs));
  if (newmsgs == 0)
    SendLedUpdate (s, 0);
  else
    SendLedUpdate (s, 1);
  return 0;
}

/*--- unistim_new: Initiate a call in the UNISTIM channel */
/*      called from unistim_request (calls from the pbx ) */
static struct ast_channel *
unistim_new (struct unistim_subchannel *sub, int state)
{
  struct ast_channel *tmp;
  struct unistim_line *l = sub->parent;
  int fmt;
  l = sub->parent;
  if (unistimdebug)
    ast_verbose ("unistim_new sub=%p\n", sub);
  tmp = ast_channel_alloc (1);
  if (tmp)
    {
      tmp->nativeformats = l->capability;
      if (!tmp->nativeformats)
	tmp->nativeformats = CAPABILITY;
      fmt = ast_best_codec (tmp->nativeformats);
      if (unistimdebug)
	ast_verbose ("best codec = %d from nativeformats %d (line cap=%d global=%d)\n", fmt, tmp->nativeformats,
		     l->capability, CAPABILITY);
      snprintf (tmp->name, sizeof (tmp->name), "USTM/%s@%s-%d", l->name, l->parent->name, sub->callid);
      if (sub->rtp)
	{
	  tmp->fds[0] = ast_rtp_fd (sub->rtp);
	  tmp->fds[1] = ast_rtcp_fd (sub->rtp);
	}
      tmp->type = type;
      ast_setstate (tmp, state);
      if (state == AST_STATE_RING)
	tmp->rings = 1;
      tmp->writeformat = fmt;
      tmp->rawwriteformat = fmt;
      tmp->readformat = fmt;
      tmp->rawreadformat = fmt;
      tmp->tech_pvt = sub;
      tmp->tech = &unistim_tech;
      if (!ast_strlen_zero (l->language))
	strncpy (tmp->language, l->language, sizeof (tmp->language) - 1);
      if (!ast_strlen_zero (l->accountcode))
	strncpy (tmp->accountcode, l->accountcode, sizeof (tmp->accountcode) - 1);
      if (l->amaflags)
	tmp->amaflags = l->amaflags;
      sub->owner = tmp;
      ast_mutex_lock (&usecnt_lock);
      usecnt++;
      ast_mutex_unlock (&usecnt_lock);
      ast_update_use_count ();
      tmp->callgroup = l->callgroup;
      tmp->pickupgroup = l->pickupgroup;
      strncpy (tmp->call_forward, l->call_forward, sizeof (tmp->call_forward) - 1);
      strncpy (tmp->context, l->context, sizeof (tmp->context) - 1);
      strncpy (tmp->exten, l->exten, sizeof (tmp->exten) - 1);
      if (!ast_strlen_zero (l->cid_num))
	{
	  tmp->cid.cid_num = strdup (l->cid_num);
	}
      tmp->priority = 1;
      if (state != AST_STATE_DOWN)
	{
	  if (ast_pbx_start (tmp))
	    {
	      ast_log (LOG_WARNING, "Unable to start PBX on %s\n", tmp->name);
	      ast_hangup (tmp);
	      tmp = NULL;
	    }
	}
    }
  else
    {
      ast_log (LOG_WARNING, "Unable to allocate channel structure\n");
    }
  return tmp;
}

static void *
do_monitor (void *data)
{
  struct unistimsession *cur = NULL;
  unsigned int tick;
  unsigned int dwOldTimeOut = MAXINT;
  int i;
  int res;
  int reloading;
  time_t t;
  /* Add an I/O event to our UDP socket */
  if (unistimsock > -1)
    ast_io_add (io, unistimsock, unistimsock_read, AST_IO_IN, NULL);

  /* This thread monitors all the frame relay interfaces which are not yet in use
     (and thus do not have a separate thread) indefinitely */
  /* From here on out, we die whenever asked */
  for (;;)
    {
      /* Wait for sched or io */
      // Looking for the smallest time-out value
      tick = GetTickCount ();
      dwOldTimeOut = dwTimeOut;
      dwTimeOut = MAXINT;
      ast_mutex_lock (&sessionlock);
      cur = sessions;
      DEBUG_TIMER ("cur = %p\n", cur);
      while (cur)
	{
	  DEBUG_TIMER ("cur = %p timeout %d\n", cur, cur->timeout);
	  if (cur->timeout < tick)
	    {
	      CallBackTimeOut (0);
	      i = 0;
	    }
	  if (dwTimeOut > cur->timeout - tick)
	    dwTimeOut = cur->timeout - tick;
	  cur = cur->next;
	}
      ast_mutex_unlock (&sessionlock);
      DEBUG_TIMER ("old %d new %d\n", dwOldTimeOut, dwTimeOut);
      if ((dwOldTimeOut != dwTimeOut) && (dwOldTimeOut - IDLE_WAIT - 30 > dwTimeOut) && (dwOldTimeOut - IDLE_WAIT + 30 < dwTimeOut))
	{
	  DEBUG_TIMER ("Waiting with ast_sched_wait() for a maximum of %d ms\n", dwTimeOut);
	  DEBUG_TIMER ("taskid was == %d\n", taskid);
	  if (taskid != -1)
	    ast_sched_del (sched, taskid);
	  if (dwTimeOut != MAXINT)
	    {
	      taskid = ast_sched_add (sched, dwTimeOut, CallBackTimeOut, 0);
	      if (taskid < 0)
		ast_log (LOG_WARNING, "ast_sched_add failed\n");
	      DEBUG_TIMER ("taskid is == %d with time out == %d\n", taskid, dwTimeOut);
	    }
	}
      if ((dwTimeOut != MAXINT) && (taskid == -1))
	{
	  DEBUG_TIMER ("taskid was -1 and dwTimeOut %d\n", dwTimeOut);
	  taskid = ast_sched_add (sched, dwTimeOut, CallBackTimeOut, 0);
	}
      res = dwTimeOut;
      if ((res < 0) || (res > IDLE_WAIT))
	res = 1000;
      res = ast_io_wait (io, res);
      ast_mutex_lock (&monlock);
      if (res >= 0)
	ast_sched_runq (sched);
      time (&t);
      cur = sessions;
      while (cur)
	{
	  if (cur->device)
	    {
	      if ((!ast_strlen_zero (cur->device->lines->mailbox)) && ((t - cur->device->lines->lastmsgcheck) > 10))
		{
		  unistim_send_mwi_to_peer (cur);
		  break;
		}
	    }
	  cur = cur->next;
	}
      ast_mutex_unlock (&monlock);
      /* Check for a reload request */
      ast_mutex_lock (&unistim_reload_lock);
      reloading = unistim_reloading;
      unistim_reloading = 0;
      ast_mutex_unlock (&unistim_reload_lock);
      if (reloading)
	{
	  if (option_verbose > 0)
	    ast_verbose (VERBOSE_PREFIX_1 "Reloading UNISTIM - disabled in this version\n");
	  // unistim_do_reload ();
	}
    }
  /* Never reached */
  return NULL;

}


/*--- restart_monitor: Start the channel monitor thread ---*/
static int
restart_monitor (void)
{
  pthread_attr_t attr;
  /* If we're supposed to be stopped -- stay stopped */
  if (monitor_thread == AST_PTHREADT_STOP)
    return 0;
  if (ast_mutex_lock (&monlock))
    {
      ast_log (LOG_WARNING, "Unable to lock monitor\n");
      return -1;
    }
  if (monitor_thread == pthread_self ())
    {
      ast_mutex_unlock (&monlock);
      ast_log (LOG_WARNING, "Cannot kill myself\n");
      return -1;
    }
  if (monitor_thread != AST_PTHREADT_NULL)
    {
      /* Wake up the thread */
      pthread_kill (monitor_thread, SIGURG);
    }
  else
    {
      pthread_attr_init (&attr);
      pthread_attr_setdetachstate (&attr, PTHREAD_CREATE_DETACHED);
      /* Start a new monitor */
      if (ast_pthread_create (&monitor_thread, &attr, do_monitor, NULL) < 0)
	{
	  ast_mutex_unlock (&monlock);
	  ast_log (LOG_ERROR, "Unable to start monitor thread.\n");
	  return -1;
	}
    }
  ast_mutex_unlock (&monlock);
  return 0;
}

/*--- unistim_request: PBX interface function ---*/
/* UNISTIM calls initiated by the PBX arrive here */
static struct ast_channel *
unistim_request (const char *type, int format, void *data, int *cause)
{
  int oldformat;
  struct unistim_subchannel *sub;
  struct ast_channel *tmpc = NULL;
  char tmp[256];
  char *dest = data;

  oldformat = format;
  format &= CAPABILITY;
  ast_log (LOG_NOTICE, "Asked to get a channel of format %s while capability is %d result : %s (%d) \n",
	   ast_getformatname (oldformat), CAPABILITY, ast_getformatname (format), format);
  if (!format)
    {
      ast_log (LOG_NOTICE, "Asked to get a channel of unsupported format %s while capability is %s\n",
	       ast_getformatname (oldformat), ast_getformatname (CAPABILITY));
      return NULL;
    }

  strncpy (tmp, dest, sizeof (tmp) - 1);
  if (ast_strlen_zero (tmp))
    {
      ast_log (LOG_NOTICE, "Unistim channels require a device\n");
      return NULL;
    }

  sub = find_subchannel_by_name (tmp);
  if (!sub)
    {
      ast_log (LOG_NOTICE, "No available lines on: %s\n", dest);
      return NULL;
    }

  if (option_verbose > 2)
    {
      ast_verbose (VERBOSE_PREFIX_3 "unistim_request(%s)\n", tmp);
      ast_verbose (VERBOSE_PREFIX_3 "Unistim cw: %d, dnd: %d, so: %d, sno: %d\n",
		   sub->parent->callwaiting, sub->parent->dnd, sub->owner ? 1 : 0, sub->next->owner ? 1 : 0);
    }
  /* Busy ? */
  if (sub->owner)
    {
      if (unistimdebug)
	ast_verbose ("Busy !\n");
      return NULL;
    }
  sub->parent->capability = format;
  tmpc = unistim_new (sub->owner ? sub->next : sub, AST_STATE_DOWN);
  if (!tmpc)
    ast_log (LOG_WARNING, "Unable to make channel for '%s'\n", tmp);
  if (unistimdebug)
    ast_verbose ("unistim_request owner = %p\n", sub->owner);
  restart_monitor ();

  /* and finish */
  return tmpc;
}

static int
unistim_info (int fd, int argc, char *argv[])
{
  struct unistim_device *device = devices;
  struct unistim_line *line;
  struct unistim_subchannel *sub;
  struct unistimsession *s = sessions;
  char iabuf[INET_ADDRSTRLEN];

  if (argc != 2)
    return RESULT_SHOWUSAGE;
  ast_cli (fd, "Dumping internal structures :\ndevice\n->line\n-->sub\n");
  while (device)
    {
      ast_cli (fd, "\nname=%s id=%s line=%p ha=%p sess=%p device=%p\n", device->name, device->id, device->lines,
	       device->ha, device->session, device);
      line = device->lines;
      while (line)
	{
	  ast_cli (fd, "->name=%s sub=%p exten=%s callid=%s lastcid=%s cap=%d progress=%d device=%p line=%p\n",
		   line->name, line->sub, line->exten, line->cid_num, line->lastcallerid, line->capability,
		   line->progress, line->parent, line);
	  sub = line->sub;
	  while (sub)
	    {
	      ast_cli (fd,
		       "-->callid=%d chan=%p rtp=%p line=%p progress=%d ringing=%d outgoing=%d alreadygone=%d sub=%p\n",
		       sub->callid, sub->owner, sub->rtp, sub->parent, sub->progress, sub->ringing, sub->outgoing,
		       sub->alreadygone, sub);
	      sub = sub->next;
	    }
	  line = line->next;
	}
      device = device->next;
    }
  ast_cli (fd, "\nSessions:\n");
  while (s)
    {
      ast_cli (fd, "sin=%s timeout=%d state=%d macaddr=%s device=%p session=%p\n",
	       ast_inet_ntoa (iabuf, sizeof (iabuf), s->sin.sin_addr), s->timeout, s->state, s->macaddr, s->device, s);
      s = s->next;
    }

  return RESULT_SUCCESS;
}
static int
unistim_sp (int fd, int argc, char *argv[])
{
  struct unistim_subchannel *sub;
  int i, j = 0, len;
  unsigned char c, cc;
  unsigned char tmp[256];

  if (argc != 4)
    return RESULT_SHOWUSAGE;
  if (strlen (argv[2]) < 9)
    return RESULT_SHOWUSAGE;
  len = strlen (argv[3]);
  if (len % 2)
    return RESULT_SHOWUSAGE;
  strncpy (tmp, argv[2] + 5, sizeof (tmp) - 1);
  sub = find_subchannel_by_name (tmp);
  if (!sub)
    {
      ast_cli (fd, "Can't find '%s'\n", tmp);
      return RESULT_SUCCESS;
    }
  if (!sub->parent->parent->session)
    {
      ast_cli (fd, "'%s' is not connected\n", tmp);
      return RESULT_SUCCESS;
    }
  ast_cli (fd, "Sending '%s' to %s (%p)\n", argv[3], tmp, sub->parent->parent->session);
  for (i = 0; i < len; i++)
    {
      c = argv[3][i];
      if (c >= 'a')
	c -= 'a' - 10;
      else
	c -= '0';
      i++;
      cc = argv[3][i];
      if (cc >= 'a')
	cc -= 'a' - 10;
      else
	cc -= '0';
      tmp[j++] = (c << 4) | cc;
    }
  memcpy (buffsend + SIZE_HEADER, tmp, j);
  SendClient (SIZE_HEADER + j, buffsend, sub->parent->parent->session);
  return RESULT_SUCCESS;
}

static int
unistim_do_debug (int fd, int argc, char *argv[])
{
  if (argc != 2)
    return RESULT_SHOWUSAGE;
  unistimdebug = 1;
  ast_cli (fd, "UNISTIM Debugging Enabled\n");
  return RESULT_SUCCESS;
}

static int
unistim_no_debug (int fd, int argc, char *argv[])
{
  if (argc != 3)
    return RESULT_SHOWUSAGE;
  unistimdebug = 0;
  ast_cli (fd, "UNISTIM Debugging Disabled\n");
  return RESULT_SUCCESS;
}

static char info_usage[] = "Usage: unistim info\n" "       Dump internal structures.\n";

static char debug_usage[] = "Usage: unistim debug\n" "       Enables dumping of UNISTIM packets for debugging purposes\n";

static char no_debug_usage[] = "Usage: unistim no debug\n" "       Disables dumping of UNISTIM packets for debugging purposes\n";
static char unistim_reload_usage[] = "Usage: unistim reload\n" "       Reloads UNISTIM configuration from unistim.conf\n";

static char sp_usage[] = "Usage: unistim sp USTM/line@name hexa\n" "       unistim sp USTM/1000@hans 19040004\n";
static struct ast_cli_entry cli_info = { {"unistim", "info", NULL}, unistim_info, "Show UNISTIM info",
info_usage
};
static struct ast_cli_entry cli_sp = { {"unistim", "sp", NULL}, unistim_sp, "Send packet (for reverse engineering)", sp_usage };
static struct ast_cli_entry cli_debug = { {"unistim", "debug", NULL}, unistim_do_debug, "Enable UNISTIM debugging",
debug_usage
};
static struct ast_cli_entry cli_no_debug = { {"unistim", "no", "debug", NULL}, unistim_no_debug,
"Disable UNISTIM debugging", no_debug_usage
};

static void
unquote (char *out, char *src, int maxlen)
{
  int len = strlen (src);
  if (!len)
    return;
  if ((len > 1) && src[0] == '\"')
    {
      /* This is a quoted string */
      src++;
      /* Don't take more than what's there */
      len--;
      if (maxlen > len - 1)
	maxlen = len - 1;
      memcpy (out, src, maxlen);
      ((char *) out)[maxlen] = '\0';
    }
  else
    memcpy (out, src, maxlen);
  return;
}

static void
ParseBookmark (char *text, struct unistim_device *d, int *pos)
{
  char line[256];
  char *at;
  char *number;
  char *icon;

  strncpy (line, text, sizeof (line) - 1);
  at = strchr (line, '@');
  if (!at)
    {
      ast_log (LOG_NOTICE, "Bookmark entry '%s' has no @ (at) sign!\n", text);
      return;
    }
  *at = '\0';
  at++;
  number = at;
  at = strchr (at, '@');
  if (!strlen (number))
    {
      ast_log (LOG_NOTICE, "Bookmark entry '%s' has no number\n", text);
      return;
    }
  if (!strlen (line))
    {
      ast_log (LOG_NOTICE, "Bookmark entry '%s' has no description\n", text);
      return;
    }

  at = strchr (number, '@');
  if (!at)
    d->softkeyicon[*pos] = FAV_ICON_SHARP;	/* default icon */
  else
    {
      *at = '\0';
      at++;
      icon = at;
      if (!strlen (icon))
	{
	  ast_log (LOG_NOTICE, "Bookmark entry '%s' has no icon value\n", text);
	  return;
	}
      if (strncmp (icon, "USTM/", 5))
	d->softkeyicon[*pos] = atoi (icon);
      else
	{
	  d->softkeyicon[*pos] = 1;
	  strncpy (d->softkeydevice[*pos], icon + 5, sizeof (d->softkeydevice[*pos]));
	}
    }
  strncpy (d->softkeylabel[*pos], line, sizeof (d->softkeylabel[*pos]) - 1);
  strncpy (d->softkeynumber[*pos], number, sizeof (d->softkeynumber[*pos]) - 1);
  if (unistimdebug)
    ast_verbose ("New bookmark at pos %d label='%s' number='%s' icon=%x\n",
		 *pos, d->softkeylabel[*pos], d->softkeynumber[*pos], d->softkeyicon[*pos]);
  (*pos)++;
}

/* Looking for dynamic icons entries in bookmarks */
static void
FinishBookmark ()
{
  struct unistim_device *d = devices;
  int i;

  while (d)
    {
      for (i = 0; i < 6; i++)
	{
	  if (d->softkeyicon[i] == 1)	/* Something for us */
	    {
	      struct unistim_device *d2 = devices;
	      while (d2)
		{
		  ast_verbose ("comp '%s' and '%s'\n", d->softkeydevice[i], d2->name);
		  if (!strcmp (d->softkeydevice[i], d2->name))
		    {
		      d->sp[i] = d2;
		      d->softkeyicon[i] = 0;
		      break;
		    }
		  d2 = d2->next;
		}
	      if (d->sp[i] == NULL)
		ast_log (LOG_NOTICE, "Bookmark entry with device %s not found", d->buff_entry);
	    }
	}
      d = d->next;
    }
}

static struct unistim_device *
build_device (char *cat, struct ast_variable *v)
{
  struct unistim_device *d;
  struct unistim_line *l;
  struct unistim_subchannel *sub;
  int i = 0, y = 0, pos, dateformat, timeformat;
  char country[3];
  char ringvolume, ringstyle;

  d = malloc (sizeof (struct unistim_device));
  if (d)
    {
      memset (d, 0, sizeof (struct unistim_device));
      strncpy (d->name, cat, sizeof (d->name) - 1);
      d->contrast = -1;
      d->output = OUTPUT_HANDSET;
      d->previous_output = OUTPUT_HANDSET;
      d->volume = VOLUME_LOW;
      d->mute = MUTE_OFF;
      linelabel[0] = 0;
      dateformat = 1;
      timeformat = 1;
      ringvolume = 2;
      ringstyle = 3;
      callerid[0] = '\0';
      pos = 0;
      while (v)
	{
	  if (!strcasecmp (v->name, "rtp_port"))
	    d->rtp_port = atoi (v->value);
	  else if (!strcasecmp (v->name, "rtp_method"))
	    d->rtp_method = atoi (v->value);
	  else if (!strcasecmp (v->name, "device"))
	    strncpy (d->id, v->value, sizeof (d->id) - 1);
	  else if (!strcasecmp (v->name, "permit") || !strcasecmp (v->name, "deny"))
	    d->ha = ast_append_ha (v->name, v->value, d->ha);
	  else if (!strcasecmp (v->name, "context"))
	    strncpy (context, v->value, sizeof (context) - 1);
	  else if (!strcasecmp (v->name, "maintext0"))
	    unquote (d->maintext0, v->value, sizeof (d->maintext0) - 1);
	  else if (!strcasecmp (v->name, "maintext1"))
	    unquote (d->maintext1, v->value, sizeof (d->maintext1) - 1);
	  else if (!strcasecmp (v->name, "maintext2"))
	    unquote (d->maintext2, v->value, sizeof (d->maintext2) - 1);
	  else if (!strcasecmp (v->name, "titledefault"))
	    unquote (d->titledefault, v->value, sizeof (d->titledefault) - 1);
	  else if (!strcasecmp (v->name, "dateformat"))
	    dateformat = atoi (v->value);
	  else if (!strcasecmp (v->name, "timeformat"))
	    timeformat = atoi (v->value);
	  else if (!strcasecmp (v->name, "contrast"))
	    {
	      d->contrast = atoi (v->value);
	      if ((d->contrast < 0) || (d->contrast > 15))
		{
		  ast_log (LOG_WARNING, "constrast must be beetween 0 and 15");
		  d->contrast = 8;
		}
	    }
	  else if (!strcasecmp (v->name, "nat"))
	    nat = ast_true (v->value);
	  else if (!strcasecmp (v->name, "ringvolume"))
	    ringvolume = atoi (v->value);
	  else if (!strcasecmp (v->name, "ringstyle"))
	    ringstyle = atoi (v->value);
	  else if (!strcasecmp (v->name, "callerid"))
	    {
	      if (!strcasecmp (v->value, "asreceived"))
		callerid[0] = '\0';
	      else
		strncpy (callerid, v->value, sizeof (callerid) - 1);
	    }
	  else if (!strcasecmp (v->name, "language"))
	    strncpy (language, v->value, sizeof (language) - 1);
	  else if (!strcasecmp (v->name, "country"))
	    strncpy (country, v->value, sizeof (country) - 1);
	  else if (!strcasecmp (v->name, "accountcode"))
	    strncpy (accountcode, v->value, sizeof (accountcode) - 1);
	  else if (!strcasecmp (v->name, "amaflags"))
	    {
	      y = ast_cdr_amaflags2int (v->value);
	      if (y < 0)
		ast_log (LOG_WARNING, "Invalid AMA flags: %s at line %d\n", v->value, v->lineno);
	      else
		amaflags = y;
	    }
	  else if (!strcasecmp (v->name, "musiconhold"))
	    strncpy (musicclass, v->value, sizeof (musicclass) - 1);
	  else if (!strcasecmp (v->name, "callgroup"))
	    cur_callergroup = ast_get_group (v->value);
	  else if (!strcasecmp (v->name, "pickupgroup"))
	    cur_pickupgroup = ast_get_group (v->value);
	  else if (!strcasecmp (v->name, "immediate"))
	    immediate = ast_true (v->value);
	  else if (!strcasecmp (v->name, "cancallforward"))
	    cancallforward = ast_true (v->value);
	  else if (!strcasecmp (v->name, "mailbox"))
	    strncpy (mailbox, v->value, sizeof (mailbox) - 1);
	  else if (!strcasecmp (v->name, "callreturn"))
	    callreturn = ast_true (v->value);
	  else if (!strcasecmp (v->name, "immediate"))
	    immediate = ast_true (v->value);
	  else if (!strcasecmp (v->name, "callwaiting"))
	    callwaiting = ast_true (v->value);
	  else if (!strcasecmp (v->name, "transfer"))
	    transfer = ast_true (v->value);
	  else if (!strcasecmp (v->name, "threewaycalling"))
	    threewaycalling = ast_true (v->value);
	  else if (!strcasecmp (v->name, "linelabel"))
	    unquote (linelabel, v->value, sizeof (linelabel) - 1);
	  else if (!strcasecmp (v->name, "bookmark"))
	    {
	      if (pos > 5)
		ast_log (LOG_WARNING, "More than 6 softkeys defined. Ignoring new entries.\n");
	      else
		ParseBookmark (v->value, d, &pos);
	    }
	  else if (!strcasecmp (v->name, "line"))
	    {
	      if (pos)
		{
		  ast_log (LOG_WARNING, "You must use bookmark AFTER line=>. Only one line is supported in this version\n");
		  return NULL;
		}
	      l = malloc (sizeof (struct unistim_line));;
	      if (l)
		{
		  memset (l, 0, sizeof (struct unistim_line));
		  ast_mutex_init (&l->lock);
		  strncpy (l->name, v->value, sizeof (l->name) - 1);

		  /* XXX Should we check for uniqueness?? XXX */
		  if (!strlen (linelabel))	// label is not defined ?
		    {
		      strncpy (d->softkeylabel[pos], v->value, sizeof (d->softkeylabel[pos]) - 1);
		      d->softkeyicon[i] = FAV_ICON_ONHOOK_BLACK;
		    }
		  else
		    {
		      strncpy (d->softkeylabel[pos], linelabel, sizeof (d->softkeylabel[pos]) - 1);
		      d->softkeyicon[pos] = FAV_ICON_ONHOOK_BLACK;
		    }
		  pos++;
		  strncpy (l->context, context, sizeof (l->context) - 1);
		  strncpy (l->cid_num, callerid, sizeof (l->cid_num) - 1);
		  strncpy (l->label, linelabel, sizeof (l->label) - 1);
		  strncpy (l->language, language, sizeof (l->language) - 1);
		  strncpy (l->musicclass, musicclass, sizeof (l->musicclass) - 1);
		  strncpy (l->mailbox, mailbox, sizeof (l->mailbox) - 1);
		  strncpy (l->mailbox, mailbox, sizeof (l->mailbox) - 1);
		  if (!ast_strlen_zero (mailbox))
		    {
		      if (unistimdebug)
			ast_verbose (VERBOSE_PREFIX_3 "Setting mailbox '%s' on %s@%s\n", mailbox, d->name, l->name);
		      mailbox[0] = '\0';
		    }

		  l->msgstate = -1;
		  l->capability = CAPABILITY;
		  l->parent = d;
		  l->immediate = immediate;
		  l->callgroup = cur_callergroup;
		  l->pickupgroup = cur_pickupgroup;
		  l->callreturn = callreturn;
		  l->cancallforward = cancallforward;
		  l->callwaiting = callwaiting;
		  l->transfer = transfer;
		  l->threewaycalling = threewaycalling;
		  l->onhooktime = time (NULL);
		  l->instance = 1;

		  for (i = 0; i < MAX_SUBS; i++)
		    {
		      sub = malloc (sizeof (struct unistim_subchannel));
		      if (sub)
			{
			  if (unistimdebug)
			    ast_verbose (VERBOSE_PREFIX_3 "Allocating UNISTIM subchannel '%d' on %s@%s ptr=%p\n", i,
					 l->name, d->name, sub);
			  memset (sub, 0, sizeof (struct unistim_subchannel));
			  ast_mutex_init (&sub->lock);
			  sub->parent = l;
			  /* Make a call*ID */
			  sub->callid = callnums;
			  callnums++;
			  sub->nat = nat;
			  sub->next = l->sub;
			  l->sub = sub;
			}
		      else
			{
			  /* XXX Should find a way to clean up our memory */
			  ast_log (LOG_WARNING, "Out of memory allocating subchannel");
			  return NULL;
			}
		    }
		  l->next = d->lines;
		  d->lines = l;
		}
	      else
		{
		  /* XXX Should find a way to clean up our memory */
		  ast_log (LOG_WARNING, "Out of memory allocating line");
		  return NULL;
		}
	    }
	  else
	    ast_log (LOG_WARNING, "Don't know keyword '%s' at line %d\n", v->name, v->lineno);
	  v = v->next;
	}
      d->ringvolume = ringvolume;
      d->ringstyle = ringstyle;
      strncpy (d->country, country, sizeof (d->country) - 1);
      d->datetimeformat = 56 + (dateformat * 4);
      d->datetimeformat += timeformat;
      if (!d->lines)
	{
	  ast_log (LOG_ERROR, "An Unistim device must have at least one line!\n");
	  return NULL;
	}
    }
  if (!strlen (d->id))
    {
      ast_log (LOG_ERROR, "You must specify the mac address with device=\n");
      return NULL;
    }
  if (!d->rtp_port)
    d->rtp_port = 10000;
  if (d->contrast == -1)
    d->contrast = 8;
  if (!strlen (d->maintext0))
    strcpy (d->maintext0, "Welcome");
  if (!strlen (d->maintext1))
    strcpy (d->maintext1, d->name);
  if (!strlen (d->titledefault))
    {
      struct tm *stm;
      time_t cur_time = time (0);

      if ((stm = localtime (&cur_time)) == 0)
	{
	  DisplayLastError ("Error in localtime()");
	  strncpy (d->titledefault, "UNISTIM for*", 12);
	}
      else
	{
	  if (strlen (stm->tm_zone) < 4)
	    {
	      strcpy (d->titledefault, "TimeZone ");
	      strcat (d->titledefault, stm->tm_zone);
	    }
	  else if (strlen (stm->tm_zone) < 9)
	    {
	      strcpy (d->titledefault, "TZ ");
	      strcat (d->titledefault, stm->tm_zone);
	    }
	  else
	    strncpy (d->titledefault, stm->tm_zone, 12);
	}
    }
  return d;
}

/*--- reload_config: Re-read unistim.conf config file ---*/
static int
reload_config (void)
{
  struct ast_config *cfg;
  struct ast_variable *v;
  struct ast_hostent ahp;
  struct hostent *hp;
  struct sockaddr_in bindaddr = { 0, };
  char iabuf[INET_ADDRSTRLEN];
  char *config = "unistim.conf";
  char *cat;
  struct unistim_device *d;

  cfg = ast_config_load (config);
  /* We *must* have a config file otherwise stop immediately */
  if (!cfg)
    {
      ast_log (LOG_ERROR, "Unable to load config %s\n", config);
      return -1;
    }

  unistim_keepalive = 120;
  unistim_port = 0;
  v = ast_variable_browse (cfg, "general");
  while (v)
    {
      if (!strcasecmp (v->name, "keepalive"))
	unistim_keepalive = atoi (v->value);
      else if (!strcasecmp (v->name, "port"))
	unistim_port = atoi (v->value);
      else if (!strcasecmp (v->name, "public_ip"))
	{
	  if (strlen (v->value))
	    {
	      if (!(hp = ast_gethostbyname (v->value, &ahp)))
		ast_log (LOG_WARNING, "Invalid address: %s\n", v->value);
	      else
		{
		  memcpy (&public_ip.sin_addr, hp->h_addr, sizeof (public_ip.sin_addr));
		  public_ip.sin_family = AF_INET;
		}
	    }
	}
      v = v->next;
    }
  if ((unistim_keepalive < 10) || (unistim_keepalive > 255 - (((NB_MAX_RETRANSMIT + 1) * RETRANSMIT_TIMER) / 1000)))
    {
      ast_log (LOG_ERROR, "keepalive is invalid in %s\n", config);
      return -1;
    }
  PacketSendPing[4] = unistim_keepalive + (((NB_MAX_RETRANSMIT + 1) * RETRANSMIT_TIMER) / 1000);
  if ((unistim_port < 1) || (unistim_port > 65535))
    {
      ast_log (LOG_ERROR, "port is not set or invalid in %s\n", config);
      return -1;
    }
  unistim_keepalive *= 1000;
  /* load the device sections */
  cat = ast_category_browse (cfg, NULL);
  while (cat)
    {
      if (strcasecmp (cat, "general"))
	{
	  d = build_device (cat, ast_variable_browse (cfg, cat));
	  if (d)
	    {
	      if (option_verbose > 2)
		ast_verbose (VERBOSE_PREFIX_3 "Added device '%s'\n", d->name);
	      d->next = devices;
	      devices = d;
	    }
	}
      cat = ast_category_browse (cfg, cat);
    }
  FinishBookmark ();
  ast_config_destroy (cfg);
  bindaddr.sin_addr.s_addr = INADDR_ANY;
  bindaddr.sin_port = htons (unistim_port);
  bindaddr.sin_family = AF_INET;
  ast_mutex_lock (&netlock);
  if (unistimsock > -1)
    {
      close (unistimsock);
      unistimsock = -1;
    }
  if (unistimsock < 0)
    {
      unistimsock = socket (AF_INET, SOCK_DGRAM, 0);
      if (unistimsock < 0)
	{
	  ast_log (LOG_WARNING, "Unable to create UNISTIM socket: %s\n", strerror (errno));
	}
      else
	{
	  const int reuseFlag = 1;
	  const int pktinfoFlag = 1;
	  setsockopt (unistimsock, SOL_SOCKET, SO_REUSEADDR, (const char *) &reuseFlag, sizeof (reuseFlag));
	  // IP_PKTINFO may not be portable. It's used to retreive our IP.
	  setsockopt (unistimsock, IPPROTO_IP, IP_PKTINFO, &pktinfoFlag, sizeof (pktinfoFlag));
	  if (bind (unistimsock, (struct sockaddr *) &bindaddr, sizeof (bindaddr)) < 0)
	    {
	      ast_log (LOG_WARNING, "Failed to bind to %s:%d: %s\n",
		       ast_inet_ntoa (iabuf, sizeof (iabuf), bindaddr.sin_addr), htons (bindaddr.sin_port), strerror (errno));
	      close (unistimsock);
	      unistimsock = -1;
	    }
	  else
	    {
	      if (option_verbose > 1)
		{
		  ast_verbose (VERBOSE_PREFIX_2
			       "UNISTIM Listening on %s:%d\n",
			       ast_inet_ntoa (iabuf, sizeof (iabuf), bindaddr.sin_addr), htons (bindaddr.sin_port));
		}
	    }
	}
    }
  ast_mutex_unlock (&netlock);
  return 0;
}
static struct ast_rtp *
unistim_get_vrtp_peer (struct ast_channel *chan)
{
  return NULL;
}

static struct ast_rtp *
unistim_get_rtp_peer (struct ast_channel *chan)
{
  struct unistim_subchannel *sub;
  if (unistimdebug)
    ast_verbose ("unistim_get_rtp_peer called\n");
  sub = chan->tech_pvt;
  if (sub && sub->rtp)
    return sub->rtp;
  return NULL;
}

static int
unistim_set_rtp_peer (struct ast_channel *chan, struct ast_rtp *rtp, struct ast_rtp *vrtp, int codecs, int nat_active)
{
  struct unistim_subchannel *sub;
  if (unistimdebug)
    ast_verbose ("unistim_set_rtp_peer called\n");
  sub = chan->tech_pvt;
  if (sub)
    {
      return 0;
    }
  return -1;
}

static struct ast_rtp_protocol unistim_rtp = {
  .type = type,
  .get_rtp_info = unistim_get_rtp_peer,
  .get_vrtp_info = unistim_get_vrtp_peer,
  .set_rtp_peer = unistim_set_rtp_peer,
};

/*--- unistim_reload: Force reload of module from cli ---*/
static int
unistim_reload (int fd, int argc, char *argv[])
{

  ast_mutex_lock (&unistim_reload_lock);
  if (unistim_reloading)
    ast_verbose ("Reload not yet done\n");
  else
    unistim_reloading = 1;
  ast_mutex_unlock (&unistim_reload_lock);
  restart_monitor ();
  return 0;
}

/*--- unistim_do_reload: Reload module 
static int
unistim_do_reload (void)
{
  reload_config ();
  return 0;
}*/

int
reload (void)
{
  return unistim_reload (0, 0, NULL);
}

static struct ast_cli_entry cli_unistim_reload = { {"unistim", "reload", NULL}, unistim_reload,
"Reload UNISTIM configuration", unistim_reload_usage
};

/*--- load_module: PBX load module - initialization ---*/
int
load_module (void)
{
  int res;
  if ((buff = malloc (SIZE_PAGE)) == 0)
    {
      ast_log (LOG_ERROR, "Error allocating network buffer\n");
      return -1;
    }
  io = io_context_create ();
  if (!io)
    {
      ast_log (LOG_WARNING, "Unable to create I/O context\n");
    }
  sched = sched_context_create ();
  if (!sched)
    {
      ast_log (LOG_WARNING, "Unable to create schedule context\n");
      free (buff);
      return -1;
    }

  res = reload_config ();
  if (!res)
    {
      /* Make sure we can register our unistim channel type */
      if (ast_channel_register (&unistim_tech))
	{
	  ast_log (LOG_ERROR, "Unable to register channel class %s\n", type);
	  unload_module ();
	  free (buff);
	  return -1;
	}

      ast_rtp_proto_register (&unistim_rtp);
      ast_cli_register (&cli_info);
      ast_cli_register (&cli_debug);
      ast_cli_register (&cli_sp);
      ast_cli_register (&cli_no_debug);
      ast_cli_register (&cli_unistim_reload);

      restart_monitor ();
    }
  return res;
}

static int
__unload_module (void)
{
  /* First, take us out of the channel loop */
  sched_context_destroy (sched);
  ast_cli_unregister (&cli_info);
  ast_cli_unregister (&cli_debug);
  ast_cli_unregister (&cli_no_debug);
  ast_cli_unregister (&cli_unistim_reload);
  ast_rtp_proto_unregister (&unistim_rtp);
  ast_channel_unregister (&unistim_tech);
  if (!ast_mutex_lock (&monlock))
    {
      if (monitor_thread && (monitor_thread != AST_PTHREADT_STOP))
	{
	  pthread_cancel (monitor_thread);
	  pthread_kill (monitor_thread, SIGURG);
	  pthread_join (monitor_thread, NULL);
	}
      monitor_thread = AST_PTHREADT_STOP;
      ast_mutex_unlock (&monlock);
    }
  else
    {
      ast_log (LOG_WARNING, "Unable to lock the monitor\n");
      return -1;
    }
  free (buff);
  return 0;
}

int
unload_module ()
{
  return __unload_module ();
}

int
usecount ()
{
  int res;
  ast_mutex_lock (&usecnt_lock);
  res = usecnt;
  ast_mutex_unlock (&usecnt_lock);
  return res;
}

char *
description ()
{
  return (char *) desc;
}


char *
key ()
{
  return ASTERISK_GPL_KEY;
}
